from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import uniswap_v2_events
from . import uniswap_v2_state

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_pool_log_deltas(
    pool: spec.Address,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    normalize: bool = True,
    include_initial_state: bool = True,
    context: spec.Context = None,
) -> spec.DataFrame:
    import asyncio
    import polars as pl

    # get start_block and initial conditions
    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )
    if start_block is None:
        start_block = await evm.async_get_contract_creation_block(
            pool, context=context
        )
        initial_point_task = None
    else:
        if include_initial_state:
            coroutine = uniswap_v2_state.async_get_pool_state(
                pool,
                block=start_block,
                normalize=normalize,
                context=context,
            )
            initial_point_task = asyncio.create_task(coroutine)
        else:
            initial_point_task = None

    # get mints, burns, and swaps
    mints_task = uniswap_v2_events.async_get_pool_mints(
        pool,
        start_block=start_block,
        end_block=end_block,
        normalize=normalize,
        context=context,
    )
    burns_task = uniswap_v2_events.async_get_pool_burns(
        pool,
        start_block=start_block,
        end_block=end_block,
        normalize=normalize,
        context=context,
    )
    swaps_task = uniswap_v2_events.async_get_pool_swaps(
        pool,
        start_block=start_block,
        end_block=end_block,
        normalize=normalize,
        context=context,
    )
    mints, burns, swaps = await asyncio.gather(
        mints_task, burns_task, swaps_task
    )

    # gather as DataFrames
    dfs = [
        pl.DataFrame(
            {
                'block_number': mints['block_number'],
                'transaction_index': mints['transaction_index'],
                'log_index': mints['log_index'],
                'event': 'Mint',
                'delta_token0': mints['arg__amount0'],
                'delta_token1': mints['arg__amount1'],
            }
        ),
        pl.DataFrame(
            {
                'block_number': mints['block_number'],
                'transaction_index': mints['transaction_index'],
                'log_index': mints['log_index'],
                'event': 'Burn',
                'delta_token0': -burns['arg__amount0'],
                'delta_token1': -burns['arg__amount1'],
            }
        ),
        pl.DataFrame(
            {
                'block_number': mints['block_number'],
                'transaction_index': mints['transaction_index'],
                'log_index': mints['log_index'],
                'event': 'Swap',
                'delta_token0': swaps['x_sold'] - swaps['x_bought'],
                'delta_token1': swaps['y_sold'] - swaps['y_bought'],
            },
        ),
    ]

    # add initial point
    if initial_point_task is not None:
        initial_point = await initial_point_task
        initial_point_df = pl.DataFrame(
            {
                'block_number': start_block,
                'transaction_index': 0,
                'log_index': 0,
                'event': 'Initial',
                'delta_token0': initial_point['x_reserves'],
                'delta_token1': initial_point['y_reserves'],
            }
        )
        dfs.append(initial_point_df)

    df = pl.concat(dfs)
    df = df.sort('block_number', 'log_index')

    return df


async def async_get_pool_transaction_deltas(
    pool: typing.Optional[spec.Address] = None,
    log_deltas: typing.Optional[spec.DataFrame] = None,
    *,
    context: spec.Context = None,
    **log_delta_kwargs: typing.Any,
) -> spec.DataFrame:

    if log_deltas is None:
        if pool is None:
            raise Exception('must specify pool or log_deltas')
        log_deltas = await async_get_pool_log_deltas(
            pool, context=context, **log_delta_kwargs
        )

    transaction_deltas: spec.DataFrame = log_deltas.groupby(
        ['block_number', 'transaction_index']
    ).sum()

    return transaction_deltas


async def async_get_pool_block_deltas(
    pool: typing.Optional[spec.Address] = None,
    log_deltas: typing.Optional[spec.DataFrame] = None,
    *,
    context: spec.Context = None,
    **log_delta_kwargs: typing.Any,
) -> spec.DataFrame:

    if log_deltas is None:
        if pool is None:
            raise Exception('must specify pool or log_deltas')
        log_deltas = await async_get_pool_log_deltas(
            pool, context=context, **log_delta_kwargs
        )

    block_deltas: spec.DataFrame = log_deltas.groupby(['block_number']).sum()

    return block_deltas


async def async_get_pool_state_per_log(
    pool: typing.Optional[spec.Address] = None,
    log_deltas: typing.Optional[spec.DataFrame] = None,
    *,
    context: spec.Context = None,
    **log_delta_kwargs: typing.Any,
) -> spec.DataFrame:

    import polars as pl

    if log_deltas is None:
        if pool is None:
            raise Exception('must specify pool or log_deltas')
        log_deltas = await async_get_pool_log_deltas(
            pool, context=context, **log_delta_kwargs
        )

    state_per_log = log_deltas.select(
        pl.col('delta_token0').cumsum().alias('token0_reserves'),
        pl.col('delta_token1').cumsum().alias('token1_reserves'),
    )

    _put_price_in_state(state_per_log)

    return state_per_log


async def async_get_pool_state_per_transaction(
    pool: typing.Optional[spec.Address] = None,
    log_deltas: typing.Optional[spec.DataFrame] = None,
    *,
    context: spec.Context = None,
    **log_delta_kwargs: typing.Any,
) -> spec.DataFrame:

    import polars as pl

    if log_deltas is None:
        if pool is None:
            raise Exception('must specify pool or log_deltas')
        log_deltas = await async_get_pool_log_deltas(
            pool, context=context, **log_delta_kwargs
        )

    transaction_deltas = await async_get_pool_transaction_deltas(
        log_deltas=log_deltas, context=context, **log_delta_kwargs
    )

    state_per_transaction = transaction_deltas.select(
        pl.col('delta_token0').cumsum().alias('token0_reserves'),
        pl.col('delta_token1').cumsum().alias('token1_reserves'),
    )

    _put_price_in_state(state_per_transaction)

    return state_per_transaction


async def async_integrate_pool_deltas(
    pool: typing.Optional[spec.Address] = None,
    *,
    interpolate: bool = False,
    log_deltas: typing.Optional[spec.DataFrame] = None,
    context: spec.Context = None,
    **log_delta_kwargs: typing.Any,
) -> spec.DataFrame:

    import polars as pl
    from ctc.toolbox import pl_utils

    if log_deltas is None:
        if pool is None:
            raise Exception('must specify pool or log_deltas')
        log_deltas = await async_get_pool_log_deltas(
            pool, context=context, **log_delta_kwargs
        )

    block_deltas = await async_get_pool_block_deltas(
        log_deltas=log_deltas, context=context, **log_delta_kwargs
    )

    state_per_block: spec.DataFrame = block_deltas.select(
        pl.col('delta_token0').cumsum().alias('token0_reserves'),
        pl.col('delta_token1').cumsum().alias('token1_reserves'),
    )

    _put_price_in_state(state_per_block)

    if interpolate:
        state_per_block = pl_utils.interpolate(
            state_per_block, index_column='block_number'
        )

    return state_per_block


def _put_price_in_state(state: spec.DataFrame) -> None:
    state['price_0_per_1'] = state['token0_reserves'] / state['token1_reserves']
    state['price_1_per_0'] = state['token1_reserves'] / state['token0_reserves']




from __future__ import annotations

import os
import typing

import polars as pl


def create_multipath_glob(paths: typing.Sequence[str]) -> str:
    """create a glob that points to a specific set of files

    works by creating a temporary directory filled with symlinks, so epheremeral
    """
    import tempfile

    temp_dir = tempfile.mkdtemp()
    for path in paths:
        os.symlink(path, os.path.join(os.path.basename(path)))
    return os.path.join(temp_dir, '*')


def write_df(
    df: pl.DataFrame | pl.LazyFrame,
    path: str,
    *,
    n_row_groups: int | None = None,
    row_group_size: int | None = None,
    statistics: bool = True,
    create_dir: bool = False,
    overwrite: bool = False,
    **write_kwargs: typing.Any,
) -> None:
    import shutil

    if create_dir:
        dirpath = os.path.dirname(path)
        if len(dirpath) > 0:
            os.makedirs(dirpath, exist_ok=True)

    if os.path.exists(path) and not overwrite:
        raise Exception('file already exists, use overwrite=True: ' + str(path))

    if n_row_groups is not None:
        if isinstance(df, pl.LazyFrame):
            raise Exception('cannot specify n_row_groups with LazyFrame\'s')
        write_kwargs['row_group_size'] = int(len(df) / n_row_groups)
        if write_kwargs['row_group_size'] == 0:
            write_kwargs['row_group_size'] = 1
    elif row_group_size is not None:
        write_kwargs['row_group_size'] = row_group_size

    temp_path = path + '_temp'
    if path.endswith('.parquet'):
        if isinstance(df, pl.LazyFrame):
            df.sink_parquet(temp_path, statistics=statistics, **write_kwargs)
        elif isinstance(df, pl.DataFrame):
            df.write_parquet(temp_path, statistics=statistics, **write_kwargs)
        else:
            raise Exception('unknown dataframe type')
    elif path.endswith('.csv'):
        if isinstance(df, pl.LazyFrame):
            df.collect(streaming=True).write_csv(temp_path, **write_kwargs)
        elif isinstance(df, pl.DataFrame):
            df.write_csv(temp_path, **write_kwargs)
        else:
            raise Exception('unknown dataframe type')
    else:
        raise Exception('unknown file format: ' + str(path))

    shutil.move(temp_path, path)


# def add_stats_to_parquet_file(
#     path: str,
#     *,
#     new_path: str | None = None,
#     replace: bool = False,
# ):
#     if new_path is None:
#         if replace:
#             new_path = path + '_tmp'
#         else:
#             pass

#     df = pl.read_parquet(path)
#     df.write_parquet(new_path, statistics=True)

#     if replace:
#         shutil.move(new_path, path)




from __future__ import annotations

from ctc import spec


_rpc_logger_state = {
    'logger_setup': False,
}


def setup_rpc_logger() -> None:
    import loguru
    from ctc import config

    if not _rpc_logger_state['logger_setup']:

        # get logging path
        rpc_log_path = config.get_rpc_requests_log_path()

        # enqueue makes logging non-blocking for async compatibility
        loguru.logger.remove()
        loguru.logger.add(
            rpc_log_path,
            enqueue=True,
            rotation='10 MB',
            format='{time} {message}',
        )

        _rpc_logger_state['logger_setup'] = True


def log_rpc_request(
    request: spec.RpcRequest, provider: spec.ProviderReference
) -> None:
    try:
        import loguru
    except ImportError:
        import warnings

        warnings.warn(
            'to use logging must install loguru, use `pip install loguru`'
        )
        return

    setup_rpc_logger()

    if isinstance(request, dict):
        loguru.logger.info(
            'request  ' + request['method'] + ' id=' + str(request['id'])
        )
    elif isinstance(request, list):
        entries = '\n'.join(
            '    ' + subrequest['method'] + ' id=' + str(subrequest['id'])
            for subrequest in request
        )
        loguru.logger.info('bulk request\n' + entries)
    else:
        raise Exception('cannot log request, unknown request type')


def log_rpc_response(
    *,
    response: spec.RpcResponse,
    request: spec.RpcRequest,
    provider: spec.ProviderReference,
) -> None:
    import loguru

    setup_rpc_logger()

    if isinstance(request, dict):
        loguru.logger.info(
            'response ' + request['method'] + ' id=' + str(request['id'])
        )
    elif isinstance(request, list):
        entries = '\n'.join(
            '    ' + subrequest['method'] + ' id=' + str(subrequest['id'])
            for subrequest in request
        )
        loguru.logger.info('bulk response\n' + entries)




from ctc.spec import typedata
from ctc.spec.typedefs import config_types
from ctc.spec.typedefs import transaction_types


def test_config_specs_match():
    assert set(config_types.Config.__annotations__.keys()) == set(
        config_types.PartialConfig.__annotations__.keys()
    )
    assert set(config_types.Config.__annotations__.keys()) == set(
        config_types.JsonConfig.__annotations__.keys()
    )

    for key, value in config_types.Config.__annotations__.items():
        assert value == config_types.PartialConfig.__annotations__[key]
        if key not in typedata.config_int_subkeys:
            assert value == config_types.JsonConfig.__annotations__[key]


def test_transaction_spec_match():
    assert set(transaction_types.DBTransaction.__annotations__.keys()) == set(
        transaction_types.DBTransactionText.__annotations__.keys()
    )




from __future__ import annotations

from ctc import spec


def get_aave_address(
    name: str,
    context: spec.Context = None,
) -> spec.Address:
    from ctc import config

    network = config.get_context_chain_id(context)
    # TODO: move to directory
    if network in ('ethereum', 1):
        if name == 'PriceOracle':
            return '0xa50ba011c48153de246e5192c8f9258a2ba79ca9'
        elif name == 'LendingPoolProvider':
            return '0xb53c1a33016b2dc2ff3653530bff1848a515c8c5'
        elif name == 'LendingPool':
            return '0x7d2768de32b0b80b7a3454c06bdac94a69ddc7a9'
        elif name == 'IncentivesController':
            return '0xd784927ff2f95ba542bfc824c8a8a98f3495f6b5'
        elif name == 'Collector':
            return '0x464C71f6c2F760DdA6093dCB91C24c39e5d6e18c'
        else:
            raise Exception('unknown contract: ' + str(name))
    else:
        raise Exception('invalid network: ' + str(network))




import pytest

from ctc import spec
from ctc.spec.typedefs import config_types
from ctc.config import config_defaults
from ctc.config import config_validate


default_db_configs = config_defaults.get_default_db_configs(
    data_dir='/path/to/data'
)

valid_values = {
    'config_spec_version': ['0.3.1'],
    'data_dir': ['/path/to/data'],
    'networks': [
        {
            1: {
                'name': 'mainnet',
                'chain_id': 1,
                'block_explorer': 'etherscan.io',
            },
        },
        {
            1: {
                'name': 'mainnet',
                'chain_id': 1,
                'block_explorer': None,
            },
        },
        {},
    ],
    'providers': [
        {
            'test_provider': {
                'name': 'test_provider',
                'url': 'http://some_url.com',
                'network': 1,
                'protocol': 'http',
                'session_kwargs': {},
                'chunk_size': None,
                'convert_reverts_to_none': False,
                'disable_batch_requests': True,
            },
        },
        {
            'test_provider': {
                'name': 'test_provider',
                'url': 'https://some_url.com',
                'network': 1,
                'protocol': 'http',
                'session_kwargs': {},
                'chunk_size': None,
                'convert_reverts_to_none': False,
                'disable_batch_requests': True,
            },
        },
    ],
    'default_network': [1],
    'default_providers': [{}],
    'db_configs': [default_db_configs],
    'log_rpc_calls': [True, False],
    'log_sql_queries': [True, False],
}

invalid_values = {
    'config_spec_version': ['0.2.10', None],
    'data_dir': ['relpath/to/data'],
    'networks': [
        {
            'mainnet': {
                'name': 'mainnet',
                'chain_id': 1,
                'block_explorer': 'etherscan.io',
            },
        },
        {
            1: {
                'name': 'mainnet',
                'chain_id': 'mainnet',
                'block_explorer': 'etherscan.io',
            },
        },
    ],
    'providers': [
        {'hi': {}},
        {
            'test_provider': {
                'name': 'test_provider',
                'url': 'some_url.com',
                'network': 1,
                'protocol': 'http',
                'session_kwargs': {},
                'chunk_size': None,
                'convert_reverts_to_none': False,
                'disable_batch_requests': True,
            },
        },
        {
            'test_provider': {
                'name': 'test_provider_different_name',
                'url': 'some_url.com',
                'network': 1,
                'protocol': 'http',
                'session_kwargs': {},
                'chunk_size': None,
                'convert_reverts_to_none': False,
                'disable_batch_requests': True,
            },
        },
    ],
    'default_network': [888, 'mainnet'],
    'default_providers': [{888: None}, {'mainnet': None}],
    'db_configs': [
        {},
    ],
    'log_rpc_calls': [None, 'a', 2],
    'log_sql_queries': [None, 'a', 2],
}


def test_every_config_key_has_validator_entry():
    config_spec = list(config_types.Config.__annotations__.keys())
    config_validators = config_validate.get_config_validators()
    for key in config_spec:
        assert key in config_validators
    for key in config_validators:
        assert key in config_spec


@pytest.mark.parametrize('item', list(valid_values.items()))
def test_validate_valid_config_values(item):
    key, values = item

    base_type = config_validate.get_config_base_types()[key]
    validator = config_validate.get_config_validators()[key]
    default_config = config_defaults.get_default_config()

    for value in values:
        assert isinstance(value, base_type)
        if validator is not None:
            validator(value, default_config)


@pytest.mark.parametrize('item', list(invalid_values.items()))
def test_validate_invalid_config_values(item):
    key, values = item

    base_type = config_validate.get_config_base_types()[key]
    validator = config_validate.get_config_validators()[key]
    default_config = config_defaults.get_default_config()

    for value in values:
        if not isinstance(value, base_type):
            continue
        if validator is not None:
            with pytest.raises(spec.ConfigInvalid):
                validator(value, default_config)




from __future__ import annotations

import os
import sys
import typing

from ctc import cli


if typing.TYPE_CHECKING:

    from typing_extensions import TypedDict

    class AliasStatus(TypedDict):
        yes_aliases: typing.Sequence[str]
        no_aliases: typing.Sequence[str]
        yes_shells: typing.Sequence[str]
        no_shells: typing.Sequence[str]


aliases_content = """
#
# # ctc aliases
#

# compute commands
alias ascii="ctc ascii"
alias hex="ctc hex"
alias keccak="ctc keccak"
alias lower="ctc lower"

# data commands
alias abi="ctc abi"
alias address="ctc address"
alias block="ctc block"
alias blocks="ctc blocks"
alias bytecode="ctc bytecode"
alias call="ctc call"
alias calls="ctc calls"
alias dex="ctc dex"
alias erc20="ctc erc20"
alias eth="ctc eth"
alias gas="ctc gas"
alias int="ctc int"
alias rlp="ctc rlp"
alias tx="ctc tx"

# protocol commands
alias 4byte="ctc 4byte"
alias aave="ctc aave"
alias cg="ctc cg"
alias chainlink="ctc chainlink"
alias curve="ctc curve"
alias es="ctc etherscan"
alias ens="ctc ens"
alias fei="ctc fei"
alias gnosis="ctc gnosis"
alias llama="ctc llama"
alias rari="ctc rari"
alias uniswap="ctc uniswap"
alias yearn="ctc yearn"

#
# # end of ctc aliases
#
"""

alias_header = """#
# # ctc aliases
#
"""

alias_footer = """
#
# # end of ctc aliases
#"""


default_shell_paths = {
    '~/.bashrc': 'bash interactive',
    '~/.profile': 'bash login',
    '~/.zshrc': 'zsh',
}


#
# # ctc root command
#


def is_ctc_on_path() -> bool:
    import subprocess

    try:
        subprocess.check_output('bash -c "which ctc"', shell=True)
        return True
    except subprocess.CalledProcessError:
        return False


def create_root_ctc_alias() -> str:
    return 'alias ctc="' + sys.executable + '-m ctc"'


#
# # alias data
#


def print_raw_aliases() -> None:
    print(aliases_content)


def get_aliases() -> typing.Mapping[str, str]:
    aliases = {}
    for line in aliases_content.split('\n'):
        if line.startswith('alias'):
            index = line.index('=')
            alias = line[6:index]
            command = line[index + 1 :]
            aliases[alias] = command
    return aliases


#
# # alias reads
#


def detect_existing_shell_paths() -> typing.Mapping[str, bool]:
    return {
        path: os.path.isfile(os.path.expanduser(path))
        for path in default_shell_paths.keys()
    }


def does_file_contain_current_aliases(path: str) -> bool:

    path = os.path.expanduser(path)

    if not os.path.isfile(path):
        return False

    with open(path, 'r') as f:
        return aliases_content in f.read()


def does_file_contain_old_aliases(path: str) -> bool:

    path = os.path.expanduser(path)

    if not os.path.isfile(path):
        return False

    with open(path, 'r') as f:
        content = f.read()
        return (
            alias_header in content
            and alias_footer in content
            and content.index(alias_header) < content.index(alias_footer)
        )


#
# # alias writes
#


def install_aliases(
    paths: typing.Sequence[str] | None = None,
    *,
    confirm: bool = False,
    verbose: bool = True,
    headless: bool = False,
) -> None:

    import toolcli

    # autodetect path based on shell
    if paths is None:
        paths = [
            path
            for path, exists in detect_existing_shell_paths().items()
            if exists
        ]

    # check that aliases are not already on path
    global_confirm = confirm
    for path in paths:
        if does_file_contain_current_aliases(path):
            if verbose:
                print('File ' + path + ' already contains shell aliases')
            continue

        elif does_file_contain_old_aliases(path):
            if verbose:
                print('File ' + path + ' contains old aliases')
            if not confirm:
                confirm = toolcli.input_yes_or_no(
                    'Replace old aliases with new aliases? ',
                    headless=headless,
                )
                if not confirm:
                    continue
                else:
                    delete_aliases(path, confirm=confirm, headless=headless)

        # confirm
        if not confirm:
            print()

            answer = toolcli.input_yes_or_no(
                'Appending ctc aliases to ' + path + '\nContinue? ',
                headless=headless,
            )
            if not answer:
                continue

        # append to file
        path = os.path.expanduser(path)
        with open(path, 'a') as f:
            f.write('\n' + aliases_content)
            if verbose:
                print('Aliases added to', path)

        confirm = global_confirm


def delete_aliases(
    path: str,
    *,
    confirm: bool = False,
    headless: bool = False,
) -> None:

    import toolcli

    path = os.path.expanduser(path)

    with open(path, 'r') as f:
        content = f.read()

    if alias_header not in content or alias_footer not in content:
        raise Exception('could not detect aliases in file ' + str(path))
    header_index = content.index(alias_header)
    footer_index = content.index(alias_footer)
    end_clip_index = footer_index + len(alias_footer)
    clipped_content = content[header_index:end_clip_index]
    new_content = content[:header_index] + content[end_clip_index:]

    # ensure not deleting anything but comments, blanks, and aliases
    clipped_lines = clipped_content.split('\n')
    for line in clipped_lines:
        valid = line.startswith('#') or line.startswith('alias ') or line == ''
        if not valid:
            raise Exception(
                'could not delete aliases, there has been addititional content placed in the ctc section of the file'
            )

    if not confirm:
        answer = toolcli.input_yes_or_no(
            'Delete aliases from file ' + path + '?',
            headless=headless,
        )
        if not answer:
            raise Exception('not deleting aliases')

    with open(path, 'w') as f:
        f.write(new_content)


#
# # alias summary
#


def get_paths_alias_status(
    paths: typing.Sequence[str] | None = None,
) -> typing.Mapping[str, typing.Literal['old', 'current', 'missing']]:

    if paths is None:
        paths = list(default_shell_paths.keys())

    statuses: typing.MutableMapping[
        str, typing.Literal['old', 'current', 'missing']
    ] = {}
    for path in paths:
        if does_file_contain_current_aliases(path):
            statuses[path] = 'current'
        elif does_file_contain_old_aliases(path):
            statuses[path] = 'old'
        else:
            statuses[path] = 'missing'
    return statuses


def print_alias_status(include_title: bool = True) -> None:
    import toolstr

    styles = cli.get_cli_styles()
    if include_title:
        toolstr.print_text_box('ctc Aliases', style=styles['title'])
    print()
    print_paths_alias_status()
    print()
    print()
    print_aliases()
    print()
    toolstr.print(
        'typing each '
        + toolstr.add_style('alias', styles['description'])
        + ' is equivalent to typing each full '
        + toolstr.add_style('command', styles['option']),
        style=styles['comment'],
    )


def print_paths_alias_status(
    paths: typing.Sequence[str] | None = None,
) -> None:

    import toolstr

    if paths is None:
        paths = list(default_shell_paths.keys())
    paths_status = get_paths_alias_status(paths)

    rows = []
    for path, status in paths_status.items():
        shell = default_shell_paths.get(path, 'unknown shell')
        row = [path, shell, status]
        rows.append(row)
    labels = ['path', 'shell', 'status']
    styles = cli.get_cli_styles()
    toolstr.print_header('Alias Configuration', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'path': styles['metavar'],
            'shell': styles['description'],
            'status': styles['description'],
        },
        indent=4,
    )

    not_current = any(status != 'current' for status in paths_status.values())
    if not_current:
        print()
        print()
        toolstr.print(
            '   install/update aliases using: '
            + toolstr.add_style('ctc aliases --install', styles['option']),
            style=styles['comment'],
        )


def print_aliases() -> None:
    import toolstr

    styles = cli.get_cli_styles()
    toolstr.print_header('Aliases', style=styles['title'])
    print()

    aliases = get_aliases()
    rows = []
    for alias, command in aliases.items():
        row = [alias, command.strip('"')]
        rows.append(row)
    labels = ['alias', 'command']
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'alias': styles['description'],
            'command': styles['option'],
        },
        indent=4,
    )



# TODO: rename this file to block_time_rpc
from __future__ import annotations

import functools
import typing

from ctc import spec
from ... import block_crud

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict, Literal

    class BlockTimestampSearchCache(TypedDict):
        initializing: dict[int, bool]
        timestamps: dict[int, int]


async def _async_get_block_of_timestamp_from_node(
    timestamp: int,
    *,
    nary: typing.Optional[int] = None,
    cache: typing.Optional[BlockTimestampSearchCache] = None,
    context: spec.Context = None,
    mode: Literal['<=', '>=', '=='] = '>=',
    verbose: bool = True,
) -> int:
    """

    - could make this efficiently parallelizable to multiple timestamps by sharing cache
        - would need to remove the initializing key from the shared cache
    """

    from ctc import config
    from ctc.toolbox import search_utils

    if nary is None:
        nary = 6

    if cache is None:
        cache = {'initializing': {timestamp: True}, 'timestamps': {}}

    async_is_match = functools.partial(
        _async_is_match_block_of_timestamp,
        timestamp=timestamp,
        cache=cache,
        context=context,
    )
    get_next_probes = functools.partial(
        _get_next_probes_block_of_timestamp,
        timestamp=timestamp,
        cache=cache,
        debug=verbose,
    )

    # determine search range
    start_index: int | None = None
    end_index: int | None = None
    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_timestamps', context=context
    )
    if read_cache:
        from ctc import db

        result = await db.async_query_timestamp_block_range(
            timestamp, context=context
        )
        if result is not None:
            start_index, end_index = result
            if start_index == end_index and start_index is not None:
                return start_index
    if start_index is None:
        start_index = 1
    if end_index is None:
        end_index = await block_crud.async_get_latest_block_number(
            context=context,
        )

    try:
        block = await search_utils.async_nary_search(
            nary=nary,
            start_index=start_index,
            end_index=end_index,
            async_is_match=async_is_match,
            get_next_probes=get_next_probes,
        )
    except search_utils.SearchRangeTooLow:
        if mode == '<=':
            return end_index
        else:
            raise Exception('no block after timestamp: ' + str(timestamp))

    if block is None:
        raise Exception('could not find block for timestamp')

    if mode == '==':
        block_data = await block_crud.async_get_block(block, context=context)
        if block_data['timestamp'] == timestamp:
            return block
        else:
            raise Exception(
                'there is no block with timestamp ' + str(timestamp)
            )
    elif mode == '<=':
        if block == 0:
            raise Exception('no block exists <= timestamp')
        block_data = await block_crud.async_get_block(block, context=context)
        if block_data['timestamp'] == timestamp:
            return block
        else:
            return block - 1
    else:
        return block


async def _async_is_match_block_of_timestamp(
    block_numbers: list[int],
    timestamp: int,
    *,
    cache: BlockTimestampSearchCache,
    context: spec.Context = None,
) -> list[bool]:

    # retrieve values not in cache
    not_in_cache = [
        block_number
        for block_number in block_numbers
        if block_number not in cache['timestamps']
    ]
    gotten = await block_crud.async_get_blocks(not_in_cache, context=context)
    for block_number, block in zip(not_in_cache, gotten):
        cache['timestamps'][block_number] = block['timestamp']

    # compute results
    return [
        cache['timestamps'][block_number] >= timestamp
        for block_number in block_numbers
    ]


def _get_next_probes_block_of_timestamp(
    *,
    nary: int,
    probe_min: int,
    probe_max: int,
    timestamp: int,
    cache: BlockTimestampSearchCache,
    debug: bool = True,
) -> list[int]:

    import numpy as np
    from ctc.toolbox import search_utils

    if cache['initializing'][timestamp]:
        cache['initializing'][timestamp] = False
        return search_utils.get_next_probes_linear(
            probe_min=probe_min, probe_max=probe_max, nary=nary
        )
    else:

        min_timestamp = cache['timestamps'][probe_min]
        total_time = cache['timestamps'][probe_max] - min_timestamp
        total_blocks = probe_max - probe_min
        mean_block_time = total_time / total_blocks
        target_index = probe_min + (timestamp - min_timestamp) / mean_block_time
        target_index = int(target_index)

        if debug:
            print('mean_block_time:', mean_block_time)
            print('probe_min:', probe_min)
            print('min_timestamp:', min_timestamp)
            print('target_index:', target_index)
            print()

        # if narrowing a range larger than N, probe a window of size X centered on target_index
        if probe_max - probe_min > 1000:
            size = probe_max - probe_min
            probes_array = target_index + size * np.linspace(
                -0.01, 0.01, nary - 1
            )
            probes_array = probes_array.astype(int)
            probes = [
                probe
                for probe in probes_array
                if probe <= probe_max and probe >= probe_min
            ]
            return probes

        # else, probe the indices directly surrounding target_index
        else:

            if target_index == probe_max:
                target_index -= 1
            elif target_index == probe_min:
                target_index += 1

            n_probes = nary - 1
            half = int(n_probes / 2)
            probes = [target_index + i for i in range(-half, n_probes - half)]

            return probes




from __future__ import annotations

import time

from ctc import evm
from ctc import spec
from .. import binary_utils
from .. import block_utils
from .. import trace_utils
from . import transaction_types


async def async_print_transaction_summary(
    transaction_hash: str,
    *,
    context: spec.Context = None,
    sort_logs_by: str | None = None,
) -> None:
    """print summary of transaction"""

    import asyncio
    import toolstr
    import tooltime
    from ctc import cli
    from ctc import rpc
    from ctc.protocols import chainlink_utils

    transaction_coroutine = rpc.async_eth_get_transaction_by_hash(
        transaction_hash,
        context=context,
    )
    transaction_receipt_task = asyncio.create_task(
        rpc.async_eth_get_transaction_receipt(
            transaction_hash=transaction_hash, context=context
        )
    )

    transaction = await transaction_coroutine
    block_task = asyncio.create_task(
        block_utils.async_get_block(
            transaction['block_number'],
            context=context,
        )
    )
    eth_usd_task = asyncio.create_task(
        chainlink_utils.async_get_eth_price(
            block=transaction['block_number'], context=context
        )
    )

    styles = cli.get_cli_styles()
    toolstr.print_text_box('Transaction Summary', style=styles['title'])
    cli.print_bullet(key='hash', value=transaction['hash'])
    cli.print_bullet(key='from', value=transaction['from'])
    cli.print_bullet(key='to', value=transaction['to'])
    cli.print_bullet(key='block number', value=transaction['block_number'])
    cli.print_bullet(
        key='transaction index', value=transaction['transaction_index']
    )
    cli.print_bullet(key='nonce', value=transaction['nonce'])

    type = transaction_types.get_transaction_type(transaction)
    type_name = transaction_types.get_transaction_type_name(type)
    type_str = type_name + ' (' + str(type) + ')'
    cli.print_bullet(key='type', value=type_str)

    cli.print_bullet(key='call data length', value=len(transaction['input']))

    block, eth_usd = await asyncio.gather(block_task, eth_usd_task)
    timestamp = block['timestamp']
    cli.print_bullet(key='timestamp', value=timestamp)
    cli.print_bullet(key='time', value=tooltime.timestamp_to_iso(timestamp))
    age = int(time.time()) - timestamp
    cli.print_bullet(key='age', value=tooltime.timelength_to_phrase(age))
    print()
    print()
    transaction_receipt = await transaction_receipt_task
    toolstr.print_text_box('Transaction Receipt', style=styles['title'])
    cli.print_bullet(key='success', value=bool(transaction_receipt['status']))
    cli.print_bullet(
        key='transaction index',
        value=str(transaction_receipt['transaction_index']),
    )
    cli.print_bullet(
        key='gas used',
        value=toolstr.format(transaction_receipt['gas_used'])
        + ' / '
        + toolstr.format(transaction['gas']),
    )
    cli.print_bullet(
        key='gas price',
        value=toolstr.format(transaction['gas_price'] / 1e9) + ' gwei',
    )
    if 'max_priority_fee_per_gas' in transaction:
        base_fee_per_gas = block.get('base_fee_per_gas')
        if base_fee_per_gas is None:
            base_fee_per_gas = 0
        cli.print_bullet(
            key='priority + base',
            value=toolstr.format(transaction['max_priority_fee_per_gas'] / 1e9)
            + ' + '
            + toolstr.format(int(base_fee_per_gas) / 1e9)
            + ' gwei',
        )
    fee = transaction_receipt['gas_used'] * transaction['gas_price'] / 1e18
    fee_usd = fee * eth_usd
    cli.print_bullet(
        key='total fee',
        value=toolstr.format(fee) + ' ETH ($' + toolstr.format(fee_usd) + ')',
    )

    if transaction['input'] == '0x':
        print()
        print()
        toolstr.print_text_box('Call Data', style=styles['title'])
        print('[none]')
    else:
        try:
            contract_abi_task = asyncio.create_task(
                evm.async_get_contract_abi(
                    contract_address=transaction['to'],
                    verbose=False,
                    context=context,
                )
            )
            contract_abi = await contract_abi_task
        except spec.AbiNotFoundException:
            print()
            print()
            print('[no contract ABI available]')
            return

        try:
            function_abi = await evm.async_get_function_abi(
                contract_address=transaction['to'],
                function_selector=transaction['input'][:10],
                context=context,
            )
        except Exception:
            print()
            print()
            print('could not find function ABI. custom proxy being used?')
            return

        call_data = evm.decode_call_data(
            call_data=transaction['input'],
            function_abi=function_abi,
        )
        print()
        print()

        from ctc.cli.commands.compute import decode_call_command

        await decode_call_command.async_decode_call_command(
            args=[transaction_hash],
            title='Call Data',
            context=context,
        )

    print()
    print()
    toolstr.print_text_box('Logs', style=styles['title'])
    logs = transaction_receipt['logs']
    if len(logs) == 0:
        print('[none]')

    else:
        # if sort_logs_by == 'signature':
        #     logs = sorted(
        #         logs,
        #         key=lambda log: binary_utils.get_event_signature(
        #             contract_address=log['address'],
        #             event_hash=log['topics'][0],
        #         ),
        #     )
        for li, log in enumerate(logs):
            if li != 0:
                print()

            event_abi = await evm.async_get_event_abi(
                contract_address=log['address'],
                event_hash=log['topics'][0],
                context=context,
            )
            normalized_event = evm.normalize_event(
                event=log,
                arg_prefix=None,
                event_abi=event_abi,
            )
            # event_signature = binary_utils.get_event_signature(event_abi=event_abi)
            stylized_event_signature = (
                toolstr.add_style(event_abi['name'], styles['option'])
                + toolstr.add_style('(', styles['title'])
                + toolstr.add_style(',', styles['title']).join(
                    toolstr.add_style(item['type'], styles['option'])
                    for item in event_abi['inputs']
                )
                + toolstr.add_style(')', styles['title'])
            )
            toolstr.print(
                stylized_event_signature,
                toolstr.add_style('-->', styles['comment']),
                toolstr.add_style(log['address'], styles['metavar']),
                style=styles['description'] + ' bold',
            )
            for e, (name, value) in enumerate(normalized_event['args'].items()):
                if (
                    value
                    == 115792089237316195423570985008687907853269984665640564039457584007913129639935
                ):
                    value = 'INT_MAX'

                # not sure if this is what should be done
                if isinstance(value, bytes):
                    value = binary_utils.to_hex(value)

                cli.print_bullet(
                    key=name,
                    value=value,
                    colon_str=' = ',
                    number=e + 1,
                    indent=4,
                )

    print()
    await trace_utils.async_print_transaction_balance_diffs(
        transaction_hash,
        styles=styles,
        context=context,
    )
    print()
    await trace_utils.async_print_transaction_storage_diffs(
        transaction_hash,
        styles=styles,
        context=context,
    )




from __future__ import annotations

import toolcli

from .. import llama_tvls


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_chains_command,
        'help': 'output data about chains tracked by Defi Llama',
        'args': [
            {
                'name': '-n',
                'type': int,
                'default': 15,
                'help': 'number of chains to display',
            },
        ],
        'examples': [
            '',
            '-n 50',
        ],
    }


async def async_llama_chains_command(n: int = 15) -> None:
    await llama_tvls.async_print_chains_tvls(n=n)




from __future__ import annotations

import typing

from ctc import spec
from . import erc4626_normalize
from . import erc4626_spec


#
# # conversions
#


async def async_convert_to_erc4626_assets(
    token: spec.Address,
    shares: typing.SupportsInt,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """convert ERC-4626 vault shares to assets"""
    from ctc import rpc

    assets: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['convertToAssets'],
        function_parameters=[int(shares)],
        block_number=block,
        context=context,
    )
    return assets


async def async_convert_to_erc4626_assets_by_block(
    token: spec.Address,
    shares: typing.SupportsInt,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """convert ERC-4626 vault shares to assets"""
    from ctc import rpc

    assets: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['convertToAssets'],
        function_parameters=[int(shares)],
        block_numbers=blocks,
        context=context,
    )
    return assets


async def async_convert_to_erc4626s_assets(
    tokens: typing.Sequence[spec.Address],
    shares: typing.Sequence[typing.SupportsInt],
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """convert ERC-4626 vault shares to assets"""

    import asyncio

    coroutines = [
        async_convert_to_erc4626_assets(
            token=token, shares=subshares, context=context, block=block
        )
        for token, subshares in zip(tokens, shares)
    ]
    return await asyncio.gather(*coroutines)


async def async_convert_to_erc4626_shares(
    token: spec.Address,
    assets: typing.SupportsInt,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """convert ERC-4626 vault assets to shares"""

    from ctc import rpc

    shares: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['convertToShares'],
        function_parameters=[int(assets)],
        block_number=block,
        context=context,
    )
    return shares


async def async_convert_to_erc4626_shares_by_block(
    token: spec.Address,
    assets: typing.SupportsInt,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """convert ERC-4626 vault assets to shares"""

    from ctc import rpc

    shares: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['convertToShares'],
        function_parameters=[int(assets)],
        block_numbers=blocks,
        context=context,
    )
    return shares


async def async_convert_to_erc4626s_shares(
    tokens: typing.Sequence[spec.Address],
    assets: typing.Sequence[typing.SupportsInt],
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """convert ERC-4626 vault assets to shares"""

    import asyncio

    coroutines = [
        async_convert_to_erc4626_shares(
            token=token, assets=subassets, context=context, block=block
        )
        for token, subassets in zip(tokens, assets)
    ]
    return await asyncio.gather(*coroutines)


#
# # max deposits, max mints, max redeems, max withdrawals
#


async def async_get_erc4626_max_deposit(
    token: spec.Address,
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> int | float:
    """get max amount of assets that receiver can deposit in ERC-4626 vault"""

    from ctc import rpc

    max_deposit: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxDeposit'],
        function_parameters=[receiver],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=max_deposit,
            context=context,
            block=block,
        )
    else:
        return max_deposit


async def async_get_erc4626_max_deposit_by_block(
    token: spec.Address,
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of assets that receiver can deposit in ERC-4626 vault"""

    from ctc import rpc

    max_deposits: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxDeposit'],
        function_parameters=[receiver],
        block_numbers=blocks,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=max_deposits,
            context=context,
            block=blocks[-1],
        )
    else:
        return max_deposits


async def async_get_erc4626s_max_deposits(
    tokens: typing.Sequence[spec.Address],
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of assets that receiver can deposit in ERC-4626 vaults"""

    from ctc import rpc

    max_deposits: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['maxDeposit'],
        function_parameters=[receiver],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626s_assets(
            tokens=tokens,
            assets=max_deposits,
            context=context,
            block=block,
        )
    else:
        return max_deposits


async def async_get_erc4626_max_mint(
    token: spec.Address,
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> int | float:
    """get max amount of shares that receiver can mint in ERC-4626 vault"""

    from ctc import rpc

    max_mint: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxMint'],
        function_parameters=[receiver],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=max_mint,
            context=context,
            block=block,
        )
    else:
        return max_mint


async def async_get_erc4626_max_mint_by_block(
    token: spec.Address,
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that receiver can mint in ERC-4626 vault"""

    from ctc import rpc

    max_mints: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxMint'],
        function_parameters=[receiver],
        block_numbers=blocks,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=max_mints,
            context=context,
            block=blocks[-1],
        )
    else:
        return max_mints


async def async_get_erc4626s_max_mints(
    tokens: typing.Sequence[spec.Address],
    receiver: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that receiver can mint in ERC-4626 vaults"""

    from ctc import rpc

    max_mints: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['maxMint'],
        function_parameters=[receiver],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626s_shares(
            tokens=tokens,
            shares=max_mints,
            context=context,
            block=block,
        )
    else:
        return max_mints


async def async_get_erc4626_max_redeem(
    token: spec.Address,
    owner: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> int | float:
    """get max amount of shares that owner can redeem from ERC-4626 vault"""

    from ctc import rpc

    max_redeem: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxRedeem'],
        function_parameters=[owner],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=max_redeem,
            context=context,
            block=block,
        )
    else:
        return max_redeem


async def async_get_erc4626_max_redeem_by_block(
    token: spec.Address,
    owner: spec.Address,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that owner can redeem from ERC-4626 vault"""

    from ctc import rpc

    max_redeems: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxRedeem'],
        function_parameters=[owner],
        block_numbers=blocks,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_shares(
            token=token,
            shares=max_redeems,
            context=context,
            block=blocks[-1],
        )
    else:
        return max_redeems


async def async_get_erc4626s_max_redeems(
    tokens: typing.Sequence[spec.Address],
    owner: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that owner can redeem from ERC-4626 vaults"""

    from ctc import rpc

    max_redeems: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['maxRedeem'],
        function_parameters=[owner],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626s_shares(
            tokens=tokens,
            shares=max_redeems,
            context=context,
            block=block,
        )
    else:
        return max_redeems


async def async_get_erc4626_max_withdraw(
    token: spec.Address,
    owner: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> int | float:
    """get max amount of shares that owner can withdraw from ERC-4626 vault"""

    from ctc import rpc

    max_withdraw: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxWithdraw'],
        function_parameters=[owner],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=max_withdraw,
            context=context,
            block=block,
        )
    else:
        return max_withdraw


async def async_get_erc4626_max_withdraw_by_block(
    token: spec.Address,
    owner: spec.Address,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that owner can withdraw from ERC-4626 vault"""

    from ctc import rpc

    max_withdraws: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['maxWithdraw'],
        function_parameters=[owner],
        block_numbers=blocks,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=max_withdraws,
            context=context,
            block=blocks[-1],
        )
    else:
        return max_withdraws


async def async_get_erc4626s_max_withdraws(
    tokens: typing.Sequence[spec.Address],
    owner: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """get max amount of shares that owner can withdraw from ERC-4626 vaults"""

    from ctc import rpc

    max_withdraws: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['maxWithdraw'],
        function_parameters=[owner],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626s_assets(
            tokens=tokens,
            assets=max_withdraws,
            context=context,
            block=block,
        )
    else:
        return max_withdraws


#
# # previews of deposits, mints, redeems, and withdrawals
#


async def async_preview_erc4626_deposit(
    token: spec.Address,
    assets: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """return shares received for assets deposited into ERC-4626 vault"""
    from ctc import rpc

    deposit: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewDeposit'],
        function_parameters=[assets],
        block_number=block,
        context=context,
    )
    return deposit


async def async_preview_erc4626_deposit_by_block(
    token: spec.Address,
    assets: float | int,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """return shares received for assets deposited into ERC-4626 vault"""
    from ctc import rpc

    deposit: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewDeposit'],
        function_parameters=[assets],
        block_numbers=blocks,
        context=context,
    )
    return deposit


async def async_preview_erc4626s_deposits(
    tokens: typing.Sequence[spec.Address],
    assets: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """return shares received for assets deposited into ERC-4626 vaults"""
    from ctc import rpc

    deposits: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['previewDeposit'],
        function_parameters=[assets],
        block_number=block,
        context=context,
    )
    return deposits


async def async_preview_erc4626_mint(
    token: spec.Address,
    shares: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """return assets needed for shared minted from ERC-4626 vault"""
    from ctc import rpc

    mint: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewMint'],
        function_parameters=[shares],
        block_number=block,
        context=context,
    )
    return mint


async def async_preview_erc4626_mint_by_block(
    token: spec.Address,
    shares: float | int,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """return assets needed for shared minted from ERC-4626 vault"""
    from ctc import rpc

    mint: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewMint'],
        function_parameters=[shares],
        block_numbers=blocks,
        context=context,
    )
    return mint


async def async_preview_erc4626s_mints(
    tokens: typing.Sequence[spec.Address],
    shares: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """return assets needed for shared minted from ERC-4626 vaults"""
    from ctc import rpc

    mints: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['previewMint'],
        function_parameters=[shares],
        block_number=block,
        context=context,
    )
    return mints


async def async_preview_erc4626_redeem(
    token: spec.Address,
    shares: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """return assets received for redeeming shares of ERC-4626 vault"""
    from ctc import rpc

    redeem: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewRedeem'],
        function_parameters=[shares],
        block_number=block,
        context=context,
    )
    return redeem


async def async_preview_erc4626_redeem_by_block(
    token: spec.Address,
    shares: float | int,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """return assets received for redeeming shares of ERC-4626 vault"""
    from ctc import rpc

    redeem: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewRedeem'],
        function_parameters=[shares],
        block_numbers=blocks,
        context=context,
    )
    return redeem


async def async_preview_erc4626s_redeems(
    tokens: typing.Sequence[spec.Address],
    shares: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """return assets received for redeeming shares of ERC-4626 vaults"""
    from ctc import rpc

    redeems: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['previewRedeem'],
        function_parameters=[shares],
        block_number=block,
        context=context,
    )
    return redeems


async def async_preview_erc4626_withdraw(
    token: spec.Address,
    assets: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    """return shares needed for redeeming assets from ERC-4626 vault"""
    from ctc import rpc

    withdraw: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewWithdraw'],
        function_parameters=[assets],
        block_number=block,
        context=context,
    )
    return withdraw


async def async_preview_erc4626_withdraw_by_block(
    token: spec.Address,
    assets: float | int,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> typing.Sequence[int]:
    """return shares needed for redeeming assets from ERC-4626 vault"""
    from ctc import rpc

    withdraw: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['previewWithdraw'],
        function_parameters=[assets],
        block_numbers=blocks,
        context=context,
    )
    return withdraw


async def async_preview_erc4626s_withdraws(
    tokens: typing.Sequence[spec.Address],
    assets: float | int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[int]:
    """return shares needed for redeeming assets from ERC-4626 vaults"""
    from ctc import rpc

    withdraws: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['previewWithdraw'],
        function_parameters=[assets],
        block_number=block,
        context=context,
    )
    return withdraws


#
# # total assets
#


async def async_get_erc4626_total_assets(
    token: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> int | float:
    """return total amount of assets in ERC-4626 vault"""
    from ctc import rpc

    assets: int = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['totalAssets'],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=assets,
            context=context,
            block=block,
        )
    return assets


async def async_get_erc4626_total_assets_by_block(
    token: spec.Address,
    *,
    context: spec.Context = None,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """return total amount of assets in ERC-4626 vault"""
    from ctc import rpc

    assets: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc4626_spec.erc4626_function_abis['totalAssets'],
        block_numbers=blocks,
        context=context,
    )
    if normalize and len(assets) > 0:
        return await erc4626_normalize.async_normalize_erc4626_assets(
            token=token,
            assets=assets,
            context=context,
            block=blocks[-1],
        )
    return assets


async def async_get_erc4626s_total_assets(
    tokens: typing.Sequence[spec.Address],
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
) -> typing.Sequence[int | float]:
    """return total amount of assets in ERC-4626 vaults"""
    from ctc import rpc

    assets: typing.Sequence[int] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        function_abi=erc4626_spec.erc4626_function_abis['totalAssets'],
        block_number=block,
        context=context,
    )
    if normalize:
        return await erc4626_normalize.async_normalize_erc4626s_assets(
            tokens=tokens,
            assets=assets,
            context=context,
            block=block,
        )
    return assets



from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    import tooltime

    class CurvePoolData(TypedDict):
        address: spec.Address
        tokens: typing.Sequence[spec.Address]
        symbols: typing.Sequence[str]
        balances: typing.Sequence[int | float | None]


# factories
curve_deployer_eoa = '0xbabe61887f1de2713c6f97e567623453d3c79f67'
old_pool_factory = '0x0959158b6040d32d04c301a72cbfd6b39e21c9ae'
pool_factory = '0xb9fc157394af804a3578134a6585c0dc9cc990d4'
deposit_separated_factory = '0xf18056bbd320e96a48e3fbf8bc061322531aac99'
crypto_factory = '0x8f942c20d02befc377d41445793068908e2250d0'
regular_factory = '0x90e00ace148ca3b23ac1bc8c240c2a7dd9c2d7f5'

eth_address = '0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'

administrative_addresses = {
    'address_provider': '0x0000000022d53366457f9d5e68ec105046fc4383',
    'pool_registry': '0x90e00ace148ca3b23ac1bc8c240c2a7dd9c2d7f5',
    'registry_registry': '0x81c46feca27b31f3adc2b91ee4be9717d1cd3dd7',
}

creation_blocks = {
    '0x0959158b6040d32d04c301a72cbfd6b39e21c9ae': 11942404,
    '0xb9fc157394af804a3578134a6585c0dc9cc990d4': 12903979,
}


# list from curve docs and manually going through non-factory pools on main page
def get_non_factory_pools() -> typing.Mapping[str, int]:
    return {
        '0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7': 10809473,  # 3Pool
        # '0x6c3f90f043a72fa612cbac8115ee7e52bde6e490': 10809467,  # 3Pool
        # '0xfd2a8fa60abd58efe3eee34dd494cd491dc14900': 11497098,  # AAVE
        '0xdebf20617708857ebe4f679508e7b7863a8a8eee': 11497106,  # AAVE
        '0xa96a65c051bf88b4095ee1f2451c2a9d43f53ae2': 11774139,  # ankrETH
        # '0xaa17a236f2badc98ddc0cf999abb47d47fc0a6cf': 11774119,  # ankrETH
        '0x79a8c46dea5ada233abaffd40f3a0a2b1e5a4f27': 9567295,  # BUSD
        # '0xb6c057591e073249f2d9d88ba59a46cfc9b59edb': 9832521,  # BUSD
        # '0x3b3ac5386837dc563660fb6a0937dfaa5924333b': 9567293,  # BUSD
        '0xa2b47e3d5c44877cca798226b7b8118f9bfb7a56': 9554040,  # Compound
        # '0xeb21209ae4c2c9ff2a86aca31e123764a3b6bc06': 9832215,  # Compound
        # '0x845838df265dcd2c412a1dc9e959c7d08537f8a2': 9554031,  # Compound
        '0x0ce6a5ff5217e38315f87032cf90686c96627caa': 11466871,  # EURS
        # '0x194ebd173f6cdace046c53eacce9b953f28411d1': 11466859,  # EURS
        '0x4ca9b3063ec5866a4b82e437059d2c43d1be596f': 10732328,  # hBTC
        # '0xb19059ebb43466c323583928285a49f558e572fd': 10732326,  # hBTC
        '0x2dded6da1bf5dbdf597c45fcfaa3194e53ecfeaf': 11831119,  # IronBank
        # '0x5282a4ef67d9c33135340fb3289cc1711c13638c': 11831114,  # IronBank
        '0xf178c0b5bb7e7abf4e12a4838c7b7c5ba2c623c0': 11875215,  # Link
        # '0xcee60cfa923170e4f8204ae08b4fa6a3f5656f3a': 11875162,  # Link
        # '0xa50ccc70b6a011cffddf45057e39679379187287': 10041096,  # PAX
        '0x06364f10b501e868329afbc005b3492902d6c763': 10041041,  # PAX
        # '0xd905e2eaebe188fc92179b6350807d8bd91db0d8': 10041032,  # PAX
        '0x93054188d876f558f4a66b2ef1d97d16edf0895b': 10151385,  # renBTC
        # '0x49849c98ae39fff122806c06791fa73784fb3675': 10151366,  # renBTC
        '0xf9440930043eb3997fc70e1339dbb11f341de7a8': 12463576,  # rETH
        # '0x53a901d48795c58f485cbb38df08fa96a24669d5': 12463570,  # rETH
        '0xeb16ae0052ed37f479f7fe63849198df1765a733': 11772500,  # sAAVE
        # '0x02d341ccb60faaf662bc0554d13778015d1b285c': 11772478,  # sAAVE
        '0x7fc77b5c7614e1533320ea6ddc2eb61fa00a9714': 10276641,  # sBTC
        # '0x075b1bb99792c9e1041ba13afef80c91a1e70fb3': 10276544,  # sBTC
        '0xc5424b857f758e906013f3555dad202e4bdb4567': 11491884,  # sETH
        # '0xa3d87fffce63b53e0d54faa1cc983b7eb0b74a9c': 11491878,  # sETH
        '0xdc24316b9ae028f1497c275eb9192a3ea0f67022': 11592551,  # stETH
        # '0x06325440d014e39736583c165c2963ba99faf14e': 11591275,  # stETH
        # '0xfcba3e75865d2d561be8d220616520c171f12851': 9906734,  # sUSD
        '0xa5407eae9ba41422680e2e00537571bcc53efbfd': 9906598,  # sUSD
        # '0xc25a3a3b969415c80451098fa907ec722572917f': 9906564,  # sUSD
        '0x80466c64868e1ab14a1ddf27a676c3fcbe638fe5': 12521538,  # TriCrypto
        # '0x331af2e331bd619defaa5dac6c038f53fcf9f785': 12602432,  # TriCrypto
        # '0xca3d75ac011bf5ad07a98d02f18225f9bd9a6bdf': 12521516,  # TriCrypto
        # '0xac795d2c97e60df6a99ff1c814727302fd747a80': 9832413,  # USDT
        '0x52ea46506b9cc5ef470c5bf89f17dc28bb35d85c': 9456293,  # USDT
        # '0x9fc689ccada600b6df723d9e47d84d76664a1f23': 9456284,  # USDT
        # '0xbbc81d23ea2c3ec7e56d39296f0cbb648873a5d3': 9832438,  # Y
        '0x45f783cce6b7ff23b2ab2d70e416cdb7d6055f51': 9476468,  # Y
        # '0xdf5e0e81dff6faf3a7e52ba697820c5e32d806a8': 9476452,  # Y
        '0x8925d9d9b4569d737a48499def3f67baa5a144b9': 11784519,  # Yv2
        # '0x571ff5b7b346f706aa48d696a9a4a288e9bb4091': 11784041,  # Yv2
        '0xd51a44d3fae010294c616388b506acda1bfaae46': 12821148,  # tricrypto2
        '0x8301ae4fc9c624d1d396cbdaa1ed877821d7c511': 13676995,  # crveth
        '0xdcef968d416a41cdac0ed8702fac8128a64241a2': 14939588,  # fraxusdc
        '0xc25099792e9349c7dd09759744ea681c7de2cb66': 11095928,  # tbtc
        '0xadcfcf9894335dc340f6cd182afa45999f45fc44': 13854276,  # xautusd
        '0x8301ae4fc9c624d1d396cbdaa1ed877821d7c511': 13676995,  # crveth
        '0xb576491f1e6e5e62f1d8f26062ee822b40b0e0d4': 13783426,  # cvxeth
        # ['reth', '0xf9440930043eb3997fc70e1339dbb11f341de7a8'],  # NOT rocket eth
    }


function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'pool_count': {
        'inputs': [],
        'name': 'pool_count',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'pool_list': {
        'inputs': [{'name': 'arg0', 'type': 'uint256'}],
        'name': 'pool_list',
        'outputs': [{'name': '', 'type': 'address'}],
        'stateMutability': 'view',
        'type': 'function',
    },
}

event_abis = {
    'BasePoolAdded__old': {
        'anonymous': False,
        'inputs': [
            {'indexed': False, 'name': 'base_pool', 'type': 'address'},
            {'indexed': False, 'name': 'implementat', 'type': 'address'},
        ],
        'name': 'BasePoolAdded',
        'type': 'event',
    },
    'BasePoolAdded__new': {
        'anonymous': False,
        'inputs': [{'indexed': False, 'name': 'base_pool', 'type': 'address'}],
        'name': 'BasePoolAdded',
        'type': 'event',
    },
    'PlainPoolDeployed': {
        'anonymous': False,
        'inputs': [
            {'indexed': False, 'name': 'coins', 'type': 'address[4]'},
            {'indexed': False, 'name': 'A', 'type': 'uint256'},
            {'indexed': False, 'name': 'fee', 'type': 'uint256'},
            {'indexed': False, 'name': 'deployer', 'type': 'address'},
        ],
        'name': 'PlainPoolDeployed',
        'type': 'event',
    },
    'MetaPoolDeployed': {
        'anonymous': False,
        'inputs': [
            {'indexed': False, 'name': 'coin', 'type': 'address'},
            {'indexed': False, 'name': 'base_pool', 'type': 'address'},
            {'indexed': False, 'name': 'A', 'type': 'uint256'},
            {'indexed': False, 'name': 'fee', 'type': 'uint256'},
            {'indexed': False, 'name': 'deployer', 'type': 'address'},
        ],
        'name': 'MetaPoolDeployed',
        'type': 'event',
    },
}

#
# # call based
#


async def async_get_factory_pool_data(
    factory: spec.Address,
    *,
    include_balances: bool = False,
    context: spec.Context = None,
) -> list[CurvePoolData]:
    import asyncio

    n_pools = await rpc.async_eth_call(
        to_address=factory,
        function_abi=function_abis['pool_count'],
        context=context,
    )

    coroutines = [
        _async_get_pool_data(
            p, factory, include_balances=include_balances, context=context
        )
        for p in range(n_pools)
    ]

    return await asyncio.gather(*coroutines)


async def _async_get_pool_data(
    p: int,
    factory: spec.Address,
    *,
    include_balances: bool = False,
    context: spec.Context = None,
) -> CurvePoolData:
    pool = await rpc.async_eth_call(
        to_address=factory,
        function_abi=function_abis['pool_list'],
        function_parameters=[p],
        context=context,
    )

    coins = await rpc.async_eth_call(
        to_address=factory,
        function_name='get_coins',  # cannot inline because different new / old
        function_parameters=[pool],
        context=context,
    )
    coins = [coin for coin in coins if coin not in [eth_address]]

    valid_coins = [
        coin
        for coin in coins
        if coin
        not in ['0x0000000000000000000000000000000000000000', eth_address]
    ]
    symbols_result = await evm.async_get_erc20s_symbols(
        valid_coins,
        context=context,
        convert_reverts_to='<UNKNOWN>',
    )
    symbols = typing.cast(list[str], symbols_result)

    if eth_address in coins:
        index = coins.index(eth_address)
        symbols.insert(index, 'ETH')

    if include_balances:
        balances: typing.MutableSequence[
            int | float | None
        ] = await evm.async_get_erc20s_balance_of(  # type: ignore
            tokens=valid_coins,
            address=pool,
            context=context,
        )
        if eth_address in coins:
            eth_balance = await evm.async_get_eth_balance(pool, context=context)
            balances.insert(index, eth_balance)
    else:
        balances = [None for coin in coins]

    return {
        'address': pool,
        'tokens': coins,
        'symbols': symbols,
        'balances': balances,
    }


#
# # event based
#


async def async_get_base_pools(
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    factory: spec.Address | None = None,
    verbose: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:
    import asyncio
    from ctc.toolbox import pl_utils

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    if start_block is None:
        start_block = 12903979

    if factory is None:
        factory = pool_factory
    if factory == pool_factory:
        factories = [old_pool_factory, pool_factory]
    else:
        factories = [factory]

    # gather data
    coroutines = []
    for factory in factories:
        if start_block is None:
            factory_start_block = creation_blocks[factory]
        else:
            factory_start_block = start_block
        coroutine = evm.async_get_events(
            contract_address=factory,
            event_name='BasePoolAdded',
            start_block=factory_start_block,
            end_block=end_block,
            context=context,
            verbose=verbose,
        )
        coroutines.append(coroutine)
    dfs = await asyncio.gather(*coroutines)

    # TODO: add event args to empty dataframes
    non_empty_dfs = [df for df in dfs if len(df) > 0]
    if len(non_empty_dfs) > 0:
        events = pl_utils.concat(non_empty_dfs)
    else:
        events = dfs[0]

    # format data
    events = events.sort('block_number')
    events = events[['block_number', 'transaction_index', 'log_index', 'contract_address', 'transaction_hash', 'arg__base_pool']]
    events = events.rename(
        {
            'contract_address': 'factory',
            'arg__base_pool': 'pool',
        }
    )

    return events


async def async_get_plain_pools(
    *,
    factory: spec.Address | None = None,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    context: spec.Context = None,
    verbose: bool = False,
) -> spec.DataFrame:

    if start_block is None:
        start_block = 12903979

    if factory is None:
        factory = pool_factory

    events = await evm.async_get_events(
        contract_address=factory,
        event_name='PlainPoolDeployed',
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        context=context,
        verbose=verbose,
    )
    events = events[
        [
            'transaction_hash',
            'contract_address',
            'arg__coins',
            'arg__A',
            'arg__fee',
            'arg__deployer',
        ]
    ]
    events = events.rename(
        {
            'contract_address': 'factory',
            'arg__coins': 'coins',
            'arg__A': 'A',
            'arg__fee': 'fee',
            'arg__deployer': 'deployer',
        }
    )
    return events


async def async_get_meta_pools(
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    factory: spec.Address | None = None,
    context: spec.Context = None,
    verbose: bool = False,
) -> spec.DataFrame:

    from ctc.toolbox import pl_utils

    if factory is None:
        factory = pool_factory
    if factory == pool_factory:
        factories = [old_pool_factory, pool_factory]
    else:
        factories = [factory]

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    # gather data
    coroutines = []
    for factory in factories:
        if start_block is None:
            factory_start_block: spec.BlockNumberReference = creation_blocks[
                factory
            ]
        else:
            factory_start_block = start_block
        coroutine = evm.async_get_events(
            contract_address=factory,
            event_name='MetaPoolDeployed',
            start_block=factory_start_block,
            end_block=end_block,
            context=context,
            verbose=verbose,
        )
        coroutines.append(coroutine)
    dfs = await asyncio.gather(*coroutines)

    # TODO: add empty args to dataframes
    non_empty_dfs = [df for df in dfs if len(df) > 0]
    if len(non_empty_dfs) > 0:
        events = pl_utils.concat(non_empty_dfs)
    else:
        events = dfs[0]

    # format data
    events = events.sort('block_number')
    events = events[
        [
            'block_number',
            'transaction_index',
            'log_index',
            'transaction_hash',
            'contract_address',
            'arg__coin',
            'arg__base_pool',
            'arg__A',
            'arg__fee',
            'arg__deployer',
        ]
    ]
    events = events.rename(
        {
            'contract_address': 'factory',
            'arg__coin': 'coin',
            'arg__base_pool': 'base_pool',
            'arg__A': 'A',
            'arg__fee': 'fee',
            'arg__deployer': 'deployer',
        }
    )

    return events




from __future__ import annotations

import typing

from ctc import spec


async def async_get_block_trace(
    block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> spec.TraceList:
    """get list of call traces for block"""
    from ctc import rpc

    result: spec.TraceList = await rpc.async_trace_block(
        block,
        context=context,
    )
    return result


async def async_get_blocks_trace(
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.TraceList]:
    """get list of call traces for block"""
    import asyncio
    from ctc import rpc
    coroutines = [
        rpc.async_trace_block(block, context=context)
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)


async def async_get_transaction_trace(
    transaction_hash: spec.TransactionHash,
    *,
    context: spec.Context = None,
) -> spec.TraceList:
    """get list of call traces for transaction"""
    from ctc import rpc

    result: spec.TraceList = await rpc.async_trace_transaction(
        transaction_hash,
        context=context,
    )
    return result


async def async_get_transaction_state_diff(
    transaction_hash: spec.TransactionHash,
    *,
    context: spec.Context = None,
) -> spec.StateDiffTrace:
    """get state diff trace for transaction"""
    from ctc import rpc

    result: spec.TraceReplayResult = await rpc.async_trace_replay_transaction(
        transaction_hash,
        trace_type=['stateDiff'],
        context=context,
    )
    state_diff = result['state_diff']
    if state_diff is None:
        raise Exception('stateDiff could not be obtained')
    else:
        return state_diff


async def async_get_transaction_vm_trace(
    transaction_hash: spec.TransactionHash,
    *,
    context: spec.Context = None,
) -> spec.VMTrace:
    """get vm trace for transaction"""
    from ctc import rpc

    result: spec.TraceReplayResult = await rpc.async_trace_replay_transaction(
        transaction_hash,
        trace_type=['vmTrace'],
        context=context,
    )
    vm_trace = result['vm_trace']
    if vm_trace is None:
        raise Exception('vmTrace could not be obtained')
    else:
        return vm_trace




from ctc.config import config_defaults


def test_default_networks_have_unique_names():
    network_names_by_chain_id = (
        config_defaults.get_default_network_names_by_chain_id()
    )
    assert len(network_names_by_chain_id) == len(
        set(network_names_by_chain_id.values())
    )


def test_all_default_networks_have_block_explorers():
    network_names_by_chain_id = (
        config_defaults.get_default_network_names_by_chain_id()
    )
    block_explorers = config_defaults.get_default_block_explorers()
    assert set(network_names_by_chain_id.values()) == set(
        block_explorers.keys()
    )



from .rpc_block_executors_sync import *
from .rpc_log_executors_sync import *
from .rpc_mining_executors_sync import *
from .rpc_node_executors_sync import *
from .rpc_state_executors_sync import *
from .rpc_submission_executors_sync import *
from .rpc_trace_executors_sync import *
from .rpc_transaction_executors_sync import *
from .rpc_whisper_executors_sync import *




from __future__ import annotations

import typing

from ctc import config
from ctc import evm
from ctc import spec

from ... import management
from ... import intake_utils
from ..block_timestamps import block_timestamps_statements
from ..transactions import transactions_intake
from . import blocks_statements

if typing.TYPE_CHECKING:
    import toolsql


async def async_intake_block(
    *,
    db_block: spec.DBBlock | None = None,
    rpc_block: spec.RPCBlock | None = None,
    context: spec.Context,
) -> None:

    if rpc_block is not None and db_block is not None:
        raise Exception('cannot specify rpc_block and db_block')
    elif rpc_block is not None:
        await async_intake_blocks(rpc_blocks=[rpc_block], context=context)
    elif db_block is not None:
        await async_intake_blocks(db_blocks=[db_block], context=context)
    else:
        raise Exception('must specify rpc_block or db_block')


async def async_intake_blocks(
    *,
    db_blocks: typing.Sequence[spec.DBBlock] | None = None,
    rpc_blocks: typing.Sequence[spec.RPCBlock] | None = None,
    latest_block_number: int | None = None,
    blocks_db_transactions: typing.Sequence[spec.DBTransaction] | None = None,
    context: spec.Context,
) -> None:
    """intake block and extract relevant information to db tables

    under normal operation should store raw block or block timestamp, noth both
    """

    import toolsql

    # check whether to intake
    active_schemas = management.get_active_schemas()
    intake_blocks = active_schemas.get('blocks')
    intake_transactions = active_schemas.get('transactions')
    intake_block_timestamps = active_schemas.get('block_timestamps')

    # get db blocks
    if rpc_blocks is not None and db_blocks is not None:
        raise Exception('should specify only rpc_blocks or db_blocks')
    if db_blocks is None:
        if rpc_blocks is None:
            raise Exception('must specify rpc_blocks or db_blocks')
        db_blocks = [
            evm.convert_rpc_block_to_db_block(rpc_block)
            for rpc_block in rpc_blocks
        ]

    if len(db_blocks) == 0:
        return

    # filter unconfirmed blocks
    db_config = config.get_context_db_config(
        schema_name='blocks',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        filtered_db_blocks = (
            await intake_utils.async_filter_fully_confirmed_blocks(
                db_blocks,
                context=context,
                latest_block_number=latest_block_number,
                conn=conn,
            )
        )
    if len(filtered_db_blocks) == 0:
        return
    db_blocks = filtered_db_blocks
    block_numbers = [block['number'] for block in db_blocks]

    # get db_block_transactions
    if (
        intake_transactions
        and blocks_db_transactions is None
        and rpc_blocks is not None
    ):
        import asyncio

        blocks_rpc_transactions: typing.Sequence[spec.RPCTransaction] = [
            tx  # type: ignore
            for rpc_block in rpc_blocks
            for tx in rpc_block['transactions']
            if rpc_block['number'] in block_numbers
            and len(rpc_block['transactions']) > 0
            and isinstance(rpc_block['transactions'][0], dict)
        ]
        coroutines = [
            evm.async_convert_rpc_transaction_to_db_transaction(
                transaction=tx, context=context
            )
            for tx in blocks_rpc_transactions
        ]
        blocks_db_transactions = await asyncio.gather(*coroutines)

    # insert into database
    if intake_blocks or intake_transactions:
        # do not perform these inserts concurrently to prevent deadlocks
        db_config = config.get_context_db_config(
            schema_name='blocks',
            context=context,
        )
        async with toolsql.async_connect(db_config) as conn:
            if intake_blocks:
                await blocks_statements.async_upsert_blocks(
                    blocks=db_blocks,
                    conn=conn,
                    context=context,
                )
            if intake_block_timestamps:
                await _async_intake_blocks_timestamps(
                    confirmed_blocks=db_blocks,
                    context=context,
                    conn=conn,
                )
        if blocks_db_transactions is not None and intake_transactions:
            await transactions_intake.async_intake_blocks_transactions(
                block_numbers=block_numbers,
                transactions=blocks_db_transactions,
                context=context,
            )


async def _async_intake_blocks_timestamps(
    confirmed_blocks: typing.Sequence[spec.DBBlock] | None = None,
    *,
    confirmed_block_timestamps: typing.Mapping[int, int] | None = None,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    if confirmed_blocks is not None and confirmed_block_timestamps is not None:
        raise Exception('cannot specify both blocks and block_timestamps')

    # determine timestamps
    if confirmed_blocks is not None:
        confirmed_block_timestamps = {
            block['number']: block['timestamp'] for block in confirmed_blocks
        }
    else:
        raise Exception('specify either blocks or block_timestamps')

    # store in database
    if len(confirmed_block_timestamps) > 0:
        await block_timestamps_statements.async_upsert_block_timestamps(
            conn=conn,
            block_timestamps=confirmed_block_timestamps,
            context=context,
        )


#
# # singular
#


# async def async_intake_block(
#     *,
#     db_block: spec.DBBlock | None = None,
#     rpc_block: spec.RPCBlock | None = None,
#     context: spec.Context,
# ) -> None:
#     """intake block and extract relevant information to db tables

#     under normal operation should store raw block or block timestamp, noth both
#     """

#     if rpc_block is not None and db_block is not None:
#         raise Exception('should specify only rpc_block or db_block')
#     if db_block is None:
#         if rpc_block is None:
#             raise Exception('must specify rpc_block or db_block')
#         else:
#             db_block = evm.block_utils.block_crud._rpc_block_to_db_block(
#                 rpc_block
#             )
#             transactions = rpc_block['transactions']
#     else:
#         transactions = None

#     # check whether to intake
#     active_schemas = management.get_active_schemas()
#     intake_block_object = active_schemas.get('blocks')
#     intake_transactions = active_schemas.get('transactions')

#     # check that block is confirmed
#     if not await intake_utils.async_is_block_fully_confirmed(
#         block=db_block['number'], context=context
#     ):
#         return

#     # insert into databases
#     if intake_block_object or intake_transactions:
#         engine = connect_utils.create_engine(
#             schema_name='block_timestamps',
#             context=context,
#         )
#         if engine is None:
#             return

#         with engine.begin() as conn:
#             # do not perform these concurrently to prevent deadlocks
#             await blocks_statements.async_upsert_block(
#                 block=db_block,
#                 conn=conn,
#                 context=context,
#             )
#             if transactions is not None and active_schemas.get('transactions'):
#                 transactions = [
#                     tx for tx in transactions if isinstance(tx, dict)
#                 ]
#                 if len(transactions) > 0:
#                     await _async_intake_block_transactions(
#                         transactions=transactions,
#                         block_number=db_block['number'],
#                         conn=conn,
#                         context=context,
#                     )


# async def _async_intake_block_timestamp(
#     block: spec.DBBlock | None,
#     *,
#     context: spec.Context,
#     conn: toolsql.SAConnection,
#     block_number: int | None = None,
#     timestamp: int | None = None,
# ) -> None:

#     # get block_number and timestamp
#     if block_number is None or timestamp is None:
#         if block is None:
#             raise Exception('must specify block or block_number and timestamp')
#         block_number = block['number']
#         timestamp = block['timestamp']

#     # store in db
#     await block_timestamps_statements.async_upsert_block_timestamp(
#         conn=conn,
#         block_number=block_number,
#         timestamp=timestamp,
#         context=context,
#     )

#
# # plural
#


# async def async_intake_blocks(
#     blocks: typing.Sequence[spec.DBBlock | spec.RPCBlock],
#     *,
#     context: spec.Context,
#     latest_block_number: int | None = None,
# ) -> None:

#     if len(blocks) == 0:
#         return

#     # check whether schemas are active
#     active_schemas = management.get_active_schemas()
#     intake_block_objects = active_schemas.get('blocks')
#     intake_block_timestamps = active_schemas.get('block_timestamps')
#     intake_block_gases = active_schemas.get('block_gas')

#     # check which blocks are confirmed
#     confirmed_blocks = await intake_utils.async_filter_fully_confirmed_blocks(
#         blocks=blocks,
#         context=context,
#         latest_block_number=latest_block_number,
#     )

#     if len(confirmed_blocks) == 0:
#         return

#     # convert to db blocks
#     db_blocks = [
#         evm.block_utils.block_crud._rpc_block_to_db_block(block)
#         for block in confirmed_blocks
#     ]

#     if intake_block_objects or intake_block_timestamps or intake_block_gases:
#         engine = connect_utils.create_engine(
#             schema_name='block_timestamps',
#             context=context,
#         )
#         if engine is None:
#             return

#         with engine.begin() as conn:
#             # do not perform these concurrently to prevent deadlocks
#             await blocks_statements.async_upsert_blocks(
#                 conn=conn,
#                 blocks=db_blocks,
#                 context=context,
#             )
#             await _async_intake_block_timestamps(
#                 confirmed_blocks=db_blocks,
#                 context=context,
#                 conn=conn,
#             )
#         await async_intake_blocks_transactions(
#             blocks=db_blocks,
#             context=context,
#         )




"""
see
- https://eth.wiki/en/fundamentals/rlp
- https://ethereum.org/en/developers/docs/data-structures-and-encoding/rlp/
"""
from __future__ import annotations

import typing

from ctc import spec
from . import format_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

    # use recursive type once mypy#731 gets released
    # https://github.com/python/mypy/issues/731

    # RLPDecodeTypes = typing.Sequence[typing.Any] | typing.ExtendedBinaryFormat | None
    RLPDecodeTypes = typing.Any


#
# # encoder
#


@typing.overload
def rlp_encode(
    item: typing.Any,
    output_format: typing.Literal['binary'],
    *,
    str_mode: Literal['auto', 'text', 'hex'] | None = None,
) -> bytes:
    ...


@typing.overload
def rlp_encode(
    item: typing.Any,
    output_format: typing.Literal['integer'],
    *,
    str_mode: Literal['auto', 'text', 'hex'] | None = None,
) -> int:
    ...


@typing.overload
def rlp_encode(
    item: typing.Any,
    output_format: typing.Literal['prefix_hex', 'raw_hex'] = 'prefix_hex',
    *,
    str_mode: Literal['auto', 'text', 'hex'] | None = None,
) -> str:
    ...


def rlp_encode(
    item: typing.Any,
    output_format: spec.BinaryFormat = 'prefix_hex',
    *,
    str_mode: Literal['auto', 'text', 'hex'] | None = 'auto',
) -> spec.GenericBinaryData:
    """encode data as RLP

    str_mode specifies how str values should be interpreted
    """

    if isinstance(item, int):
        if item == 0:
            item = bytes()
        else:
            item = format_utils.to_binary(item)

    if isinstance(item, (tuple, list)):
        output = _rlp_encode_list(item, str_mode=str_mode)
    elif isinstance(item, bytes):
        output = _rlp_encode_bytes(item)
    elif isinstance(item, str):
        output = _rlp_encode_str(item, str_mode=str_mode)
    else:
        raise Exception('cannot rlp encode items of type ' + str(type(item)))

    return format_utils.binary_convert(output, output_format)


def _rlp_encode_bytes(data: bytes) -> bytes:
    data = format_utils.to_binary(data)

    length = len(data)
    if length == 0:
        return bytes.fromhex('80')
    elif length == 1 and data <= bytes.fromhex('7f'):
        return data
    elif length <= 55:
        prefix = 128 + length
        return format_utils.to_binary(prefix) + data
    else:
        length_as_bytes = format_utils.to_binary(length)
        prefix = 183 + len(length_as_bytes)
        return format_utils.to_binary(prefix) + length_as_bytes + data


def _rlp_encode_list(
    items: typing.Sequence[typing.Any],
    str_mode: Literal['auto', 'text', 'hex'] | None = None,
) -> bytes:
    encoded_items = [
        rlp_encode(item, str_mode=str_mode, output_format='binary')
        for item in items
    ]
    item_lengths = [len(encoded_item) for encoded_item in encoded_items]
    total_payload_length = sum(item_lengths)

    if total_payload_length <= 55:
        prefix = format_utils.to_binary(192 + total_payload_length)
        output = prefix
        for item in encoded_items:
            output = output + item
        return output

    else:
        bytes_of_length = format_utils.to_binary(total_payload_length)
        prefix_int = 247 + len(bytes_of_length)
        output = format_utils.to_binary(prefix_int) + bytes_of_length
        for item in encoded_items:
            output = output + item
        return output


def _rlp_encode_str(
    item: str,
    str_mode: Literal['auto', 'text', 'hex'] | None,
) -> bytes:
    if str_mode == 'auto':
        if item.startswith('0x'):
            str_mode = 'hex'
        else:
            str_mode = 'text'

    if str_mode == 'text':
        as_bytes = item.encode()
    elif str_mode == 'hex':
        as_bytes = format_utils.to_binary(item)
    else:
        raise Exception('unknown str mode: ' + str(str_mode))

    return _rlp_encode_bytes(as_bytes)


#
# # decoder
#


def rlp_decode(
    data: spec.Data,
    types: RLPDecodeTypes = None,
) -> typing.Any:
    """decode RLP data

    the types parameter determines how raw data will be decoded in python
    - types is one of:
        1. binary format name: prefix_hex, raw_hex, binary, integer, ascii
        2. list/tuple of binary format names, corresponding to items in rlp list
    """

    # convert any abi types to RLP types
    if types is not None:
        types = _process_rlp_types(types)

    data = format_utils.to_binary(data)
    decoded, remaining = _rlp_decode_chunk(data=data, types=types)
    if len(remaining) > 0:
        raise Exception('data contains extra bytes')
    return decoded


def _process_rlp_types(
    types: str | typing.Sequence[str],
) -> str | typing.Sequence[str]:
    allowed_types = [
        'prefix_hex',
        'raw_hex',
        'binary',
        'integer',
        'ascii',
        'bool',
    ]

    if isinstance(types, str):
        if types in allowed_types:
            return types
        elif types.startswith('int') and '[' not in types:
            return 'integer'
        elif types == 'string':
            return 'ascii'
        else:
            raise Exception('RLP type not understood: ' + str(types))
    else:
        return [_process_rlp_types(type) for type in types]  # type: ignore


def _rlp_decode_chunk(
    data: bytes,
    types: RLPDecodeTypes,
) -> tuple[typing.Any, bytes]:
    """decode next item chunk in rlp encoded data"""

    if len(data) == 0:
        raise Exception()

    first_byte = data[0]
    if first_byte <= 0xBF:
        return _rlp_decode_primitive_chunk(data, types=types)
    else:
        result = _rlp_decode_list_chunk(data, types=types)
        return result


def _rlp_decode_primitive_chunk(
    data: bytes,
    types: RLPDecodeTypes,
) -> tuple[typing.Any, bytes]:
    first_byte = data[0]

    if first_byte <= 0x7F:
        # data that is single byte less than or equal to 127

        decoded: typing.Any = first_byte.to_bytes(1, 'big')
        remaining = data[1:]

    elif first_byte <= 0xB7:
        # data with length between 0 and 55 bytes

        # get length
        length = first_byte - 0x80

        # get data
        start = 1
        end = 1 + length
        decoded = data[start:end]
        remaining = data[end:]

    elif first_byte <= 0xBF:
        # data longer than 55 bytes

        # get length length
        length_length = first_byte - 0xB7

        # get length
        start = 1
        end = 1 + length_length
        length = int.from_bytes(data[start:end], 'big')

        # get data
        start = 1 + length_length
        end = 1 + length_length + length
        decoded = data[start:end]
        remaining = data[end:]

    else:
        raise Exception('next item in data is not a primitive chunk')

    # convert
    if types is not None:
        if types == 'ascii':
            decoded = decoded.decode()
        elif types == 'bool':
            decoded = bool(decoded)
        else:
            decoded = format_utils.binary_convert(decoded, types)

    return decoded, remaining


def _rlp_decode_list_chunk(
    data: bytes,
    types: RLPDecodeTypes,
) -> tuple[list[typing.Any], bytes]:
    first_byte = data[0]

    if first_byte <= 0xBF:
        raise Exception('next item in data is not a list chunk')

    elif first_byte <= 0xF7:
        # list with total payload length between 0 and 55 bytes

        # get length
        length = first_byte - 0xC0

        # decode list
        start = 1
        end = 1 + length
        list_payload = data[start:end]

        if len(list_payload) != length:
            raise Exception('data is shorted than specified by header')

        remaining = data[end:]

    else:
        # list with payload longer than 55 bytes

        # get length length
        length_length = first_byte - 0xF7

        # get length
        start = 1
        end = 1 + length_length
        length = int.from_bytes(data[start:end], 'big')

        # decode list
        start = 1 + length_length
        end = 1 + length_length + length
        list_payload = data[start:end]

        if len(list_payload) != length:
            raise Exception('data is shorted than specified by header')

        remaining = data[end:]

    # decode each item of list payload
    output = []
    remaining_payload = list_payload
    index = 0
    while len(remaining_payload) > 0:
        # get type of item
        if isinstance(types, (list, tuple)):
            if index >= len(types):
                raise Exception(
                    'number of types does not match number of list items'
                )
            item_types = types[index]
        elif isinstance(types, str):
            item_types = types
        else:
            item_types = None

        # decode next item in list
        item, remaining_payload = _rlp_decode_chunk(
            remaining_payload,
            types=item_types,
        )
        output.append(item)
        index += 1

    return output, remaining




from __future__ import annotations

from ctc import spec
from .. import format_utils
from . import key_utils
from . import secp256k1_utils
from . import vrs_utils


def verify_signature(
    signature: spec.Signature,
    *,
    message_hash: spec.Data,
    public_key: spec.Data | None = None,
    address: spec.Data | None = None,
) -> bool:
    """verify that signature was signed using given public key"""

    # recover public key from signature
    recovered_public_key = recover_signer_public_key(
        message_hash=message_hash,
        signature=signature,
    )

    # verify a match against provided public_key and/or address
    if public_key is None and address is None:
        raise Exception('must provider public_key or address to verify')
    if public_key is not None and public_key != recovered_public_key:
        return False
    if address is not None:
        recovered_address = key_utils.public_key_to_address(
            recovered_public_key
        )
        if recovered_address != address:
            return False

    return True


def get_signature_network_id(signature: spec.Signature) -> int | None:
    """extract network id from given signature

    adapted from https://github.com/ethereum/pyethereum/blob/ecb14c937a0b6cb0a0dc4f06be3a88e6d53dcce3/ethereum/transactions.py#L93
    """

    v, r, s = vrs_utils.unpack_signature_vrs(signature)
    if r == 0 and s == 0:
        return v
    elif v in (27, 28):
        return None
    else:
        return ((v - 1) // 2) - 17


def recover_signer_public_key(
    *,
    message_hash: spec.Data,
    signature: spec.Signature,
) -> str:
    """recover signer public key from signature

    adapted from https://github.com/ethereum/pyethereum/blob/ecb14c937a0b6cb0a0dc4f06be3a88e6d53dcce3/ethereum/utils.py#L87
    """

    # ger vrs
    v, r, s = vrs_utils.unpack_signature_vrs(signature)
    v = format_utils.binary_convert(v, 'integer')
    r = format_utils.binary_convert(r, 'integer')
    s = format_utils.binary_convert(s, 'integer')

    if v >= 37:
        network_id = get_signature_network_id(signature)
        if network_id is None:
            raise Exception('could not parse valid network_id')
        v = v - network_id * 2 - 8

    # get signer
    message_hash = format_utils.to_binary(message_hash)
    x, y = secp256k1_utils.ecdsa_raw_recover(message_hash, (v, r, s))
    signer = x.to_bytes(32, byteorder='big') + y.to_bytes(32, byteorder='big')

    return format_utils.to_hex(signer)


def recover_signer_address(
    *,
    message_hash: spec.Data,
    signature: spec.Signature,
) -> spec.Address:
    """recover signer public address from signature"""

    public_key = recover_signer_public_key(
        message_hash=message_hash,
        signature=signature,
    )
    return key_utils.public_key_to_address(public_key)



from __future__ import annotations

from ctc import rpc
from ctc import spec

import typing

from .. import rari_abis


fuse_directory = '0x835482fe0532f169024d5e9410199369aad5c77e'


async def async_get_all_pools(
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> typing.Sequence[typing.Sequence[typing.Any]]:

    # TODO: convert output to dict
    result = await rpc.async_eth_call(
        to_address=fuse_directory,
        block_number=block,
        function_abi=rari_abis.pool_directory_function_abis['getAllPools'],
        context=context,
    )
    if not isinstance(result, (list, tuple)) or not all(
        isinstance(item, (list, tuple)) for item in result
    ):
        raise Exception('invalid rpc result')
    return result



from __future__ import annotations

import toolcli
import toolstr

from ctc import cli
from ctc.protocols import yearn_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_yearn_addresses_command,
        'help': 'output Yearn addresses',
        'args': [
            {
                'name': ['-v', '--verbose'],
                'help': 'output additional addresses',
                'action': 'store_true',
            },
        ],
        'examples': [
            '',
            '-v',
        ],
    }


async def async_yearn_addresses_command(verbose: bool) -> None:
    styles = cli.get_cli_styles()
    toolstr.print_text_box('Yearn Lens Addresses', style=styles['title'])
    print()
    yearn_utils.print_lens_addresses(network='ethereum')
    print()
    print()
    toolstr.print_text_box('Yearn Vault Addresses', style=styles['title'])
    print()
    api_vaults = await yearn_utils.async_get_yearn_api_vaults(
        network='ethereum'
    )
    yearn_utils.print_vault_addresses(api_vaults=api_vaults, verbose=verbose)




from __future__ import annotations

import toolcli
import toolstr
import tooltime

from ctc.protocols import rari_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_pools_command,
        'help': 'list all Rari fuse pools',
        'args': [
            {
                'name': ['--verbose'],
                'help': 'emit extra output',
                'action': 'store_true',
            },
        ],
        'examples': [
            '',
            '--verbose',
        ],
    }


async def async_pools_command(verbose: bool) -> None:
    all_pools = await rari_utils.async_get_all_pools()
    rows = []
    for p, pool in enumerate(all_pools):
        if verbose:
            row = [
                p,  # index
                pool[0],  # name
                str(pool[3]),  # creation block
                tooltime.timestamp_to_iso(pool[4]),  # creation time
                pool[2],  # comptroller
                pool[1],  # creator
            ]
            labels = [
                'index',
                'name',
                'create block',
                'create time',
                'comptroller',
                'creator',
            ]
        else:
            row = [
                p,  # index
                pool[0][:20],  # name
                tooltime.timestamp_to_iso(pool[4])[:10],  # creation time
                pool[2],  # comptroller
            ]
            labels = ['index', 'name', 'date', 'comptroller']
        rows.append(row)
    print(len(rows))
    toolstr.print_table(rows, labels=labels)



from __future__ import annotations

import toolsql


schema_versions_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'schema_versions': {
            'columns': [
                {'name': 'schema_name', 'type': 'Text', 'primary': True},
                # use chain_id==-1 for schemas that are not related to a particular chain
                {'name': 'chain_id', 'type': 'Integer', 'primary': True},
                {'name': 'version', 'type': 'Text'},
            ],
        },
    },
}



from __future__ import annotations

import math
import typing

if typing.TYPE_CHECKING:
    import aiohttp

import toolstr


url_template = 'https://api.coingecko.com/api/v3/coins/markets?vs_currency=usd&order=market_cap_desc&per_page=100&page={page}&sparkline=true&price_change_percentage=1h%2C24h%2C7d'
token_url_template = 'https://www.coingecko.com/en/coins/{name}'


async def async_get_market_data(
    n: int,
    *,
    mode: str = 'serial',
) -> typing.Sequence[typing.Mapping[typing.Any, typing.Any]]:
    import asyncio
    import aiohttp

    n_per_page = 100
    n_pages = math.ceil(n / n_per_page)

    if mode == 'concurrent':
        async with aiohttp.ClientSession() as session:
            coroutines = [async_get_page(session, p) for p in range(n_pages)]
            pages = await asyncio.gather(*coroutines)

    elif mode == 'serial':
        for p in range(n_pages):
            pages = []
            async with aiohttp.ClientSession() as session:
                page = await async_get_page(session, p)
                pages.append(page)

    else:
        raise Exception('unknown mode: ' + str(mode))

    items = [item for page in pages for item in page]

    return items[:n]


async def async_get_page(
    session: aiohttp.ClientSession, p: int
) -> typing.List[typing.Any]:
    url = url_template.format(page=p + 1)
    async with session.get(url) as response:
        page = await response.json()
        if not isinstance(page, list):
            raise Exception('bad page format')
        return page


def color_polarity(value: int | float | None) -> str:
    if value is None:
        return ''

    if value > 0:
        return '#8dc647'
    elif value < 0:
        return '#e15241'
    else:
        return 'gray'


def print_market_data(
    data: typing.Sequence[typing.Any],
    *,
    verbose: bool,
    include_links: bool = False,
    height: int | None = None,
    width: int | None = None,
    n_columns: int | None = None,
    gap: int | str | None = None,
) -> None:

    import toolcli

    if width is None:
        width = 8

    # create labels
    labels = ['token', 'price', 'Δ 1H', 'Δ 24H', 'Δ 7D', 'volume', 'mkt cap']
    if verbose:
        labels.append('7D chart')

    # create rows
    rows: list[typing.Sequence[typing.Any]] = []
    for item in data:

        row = []
        row.append(item['symbol'].upper())
        row.append(item['current_price'])

        # add price change cells
        for key in [
            'price_change_percentage_1h_in_currency',
            'price_change_percentage_24h_in_currency',
            'price_change_percentage_7d_in_currency',
        ]:
            change = item[key]
            row.append(change)

        row.append(item['total_volume'])
        row.append(item['market_cap'])

        # add sparkline
        if verbose:
            from toolstr.charts import braille_utils

            sparkline = braille_utils.create_braille_sparkline(
                data=item['sparkline_in_7d']['price'],
                # width=20,
                width=width,
                height=height,
            )

            row.append(sparkline)

        rows.append(row)

    if height is None:
        height = 1

    def get_row_color(r: int) -> str:
        if height is None:
            raise Exception('height not set')
        datum = data[int(r / height)]
        diff = (
            datum['sparkline_in_7d']['price'][-1]
            - datum['sparkline_in_7d']['price'][0]
        )
        return color_polarity(diff)

    # print table
    grey = '#BBBBBB'
    table_str = toolstr.print_multiline_table(
        rows,
        labels=labels,
        column_gap=1,
        # compact=True,
        add_row_index=True,
        separate_all_rows=False,
        max_table_width=toolcli.get_n_terminal_cols(),
        vertical_justify='center',
        border='#555555',
        label_style='#DDDDDD' + ' bold',
        column_styles={
            '': '#555555' + ' bold',
            'price': grey,
            'volume': grey,
            'mkt cap': grey,
            'Δ 1H': lambda context: 'bold ' + color_polarity(context['cell']),
            'Δ 24H': lambda context: 'bold ' + color_polarity(context['cell']),
            'Δ 7D': lambda context: 'bold ' + color_polarity(context['cell']),
            '7D chart': lambda context: 'bold ' + get_row_color(context['r']),
        },
        column_formats={
            'price': {'decimals': 2, 'trailing_zeros': True, 'prefix': '$'},
            'Δ 1H': {
                'scientific': False,
                'postfix': '%',
                'decimals': 2,
                'trailing_zeros': True,
            },
            'Δ 24H': {
                'scientific': False,
                'postfix': '%',
                'decimals': 2,
                'trailing_zeros': True,
            },
            'Δ 7D': {
                'scientific': False,
                'postfix': '%',
                'decimals': 2,
                'trailing_zeros': True,
            },
            'volume': {
                'decimals': 1,
                'trailing_zeros': True,
                'prefix': '$',
                'order_of_magnitude': True,
            },
            'mkt cap': {
                'decimals': 1,
                'trailing_zeros': True,
                'prefix': '$',
                'order_of_magnitude': True,
            },
        },
        return_str=True,
    )

    if n_columns is not None:
        if table_str is None:
            raise Exception('table_str not generated')
        columnized = toolstr.columnize(
            table_str,
            n_columns=n_columns,
            gap=gap,
            header_height=2,
        )
        toolstr.print(columnized)
    else:
        toolstr.print(table_str)



from __future__ import annotations

from ctc import evm
from ctc import spec

from .. import abi_utils
from . import address_data


async def async_print_address_summary(
    address: spec.Address,
    *,
    verbose: bool | int = True,
    max_width: int = 80,
    context: spec.Context = None,
    raw: bool = False,
) -> None:
    """print summary of address

    TODO (with more data)
    - transactions
    - transfers
    """

    import asyncio
    import toolstr
    import tooltime
    from ctc import cli
    from ctc import rpc

    styles = cli.get_cli_styles()

    eth_balance_coroutine = evm.async_get_eth_balance(address, context=context)
    transaction_count_coroutine = rpc.async_eth_get_transaction_count(
        address, context=context
    )

    # contract data
    code_coroutine = rpc.async_eth_get_code(address, context=context)

    [eth_balance, transaction_count, code] = await asyncio.gather(
        eth_balance_coroutine,
        transaction_count_coroutine,
        code_coroutine,
    )

    is_contract = code != '0x'
    if is_contract:
        address_type = 'contract'
    else:
        address_type = 'EOA'

    title = 'Address ' + address.lower()
    toolstr.print_text_box(title, style=styles['title'])
    rows = [
        (
            'checksum',
            toolstr.add_style(
                address_data.get_address_checksum(address), styles['metavar']
            ),
        ),
        ('address type', address_type),
        ('ETH balance', eth_balance),
        ('transaction count', transaction_count),
    ]
    print()

    if verbose and is_contract:
        creation_block = await evm.async_get_contract_creation_block(
            address,
            context=context,
            verbose=False,
        )
        if creation_block is not None:
            block_timestamp = await evm.async_get_block_timestamp(
                creation_block, context=context
            )
            age = tooltime.get_age(block_timestamp, 'TimelengthPhrase')
            creation_tx = await evm.async_get_contract_creation_transaction(
                address,
                context=context,
                creation_block=creation_block,
            )
            deployer = await evm.async_get_contract_deployer(
                address,
                context=context,
                creation_transaction=creation_tx,
            )
            rows += [
                ('creation tx', creation_tx),
                ('deployer', deployer),
                ('creation block', creation_block),
                ('age', age),
            ]
    toolstr.print_table(
        rows,
        border=styles['comment'],
        column_justify=['right', 'left'],
        column_styles=[styles['option'], styles['description']],
    )

    if is_contract:

        is_erc20 = await evm.async_is_erc20(address, context=context)

        if is_erc20:
            print()
            print()
            await evm.async_print_erc20_summary(
                address, include_address=False, context=context
            )

        contract_abi = await abi_utils.async_get_contract_abi(
            contract_address=address,
            context=context,
        )

        print()
        print()
        if raw:
            import json
            import toolstr

            toolstr.print_header('Raw JSON ABI')
            as_str = json.dumps(contract_abi, sort_keys=True, indent=4)
            print(as_str)
        else:
            abi_utils.print_contract_abi(
                contract_abi=contract_abi,
                max_width=max_width,
                verbose=verbose,
            )




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from .. import contract_abi_utils
from . import event_abi_parsing


def get_event_abi(
    *,
    contract_abi: spec.ContractABI,
    event_name: typing.Optional[str] = None,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
) -> spec.EventABI:
    """get event ABI from contract ABI"""

    if event_abi is not None:
        return event_abi

    if event_name is None and event_hash is None:
        raise Exception('specify event_name or event_hash')

    candidates = []
    for item in contract_abi:
        if item['type'] != 'event':
            continue
        if event_name is not None and item.get('name') != event_name:
            continue
        if event_hash is not None:
            item_hash = event_abi_parsing.get_event_hash(item)
            if item_hash != event_hash:
                continue

        candidates.append(item)

    if len(candidates) == 0:
        raise LookupError('could not find event abi')
    elif len(candidates) == 1:
        return candidates[0]
    else:
        raise Exception('found too many candidates for event abi')


async def async_get_event_abi(
    *,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    contract_address: typing.Optional[spec.Address] = None,
    contract_addresses: typing.Sequence[spec.Address] | None = None,
    event_name: typing.Optional[str] = None,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
    context: spec.Context = None,
) -> spec.EventABI:
    """get event ABI from local database or block explorer"""

    if contract_addresses is not None:
        for contract_address in contract_addresses:
            try:
                return await async_get_event_abi(
                    contract_abi=contract_abi,
                    contract_address=contract_address,
                    event_name=event_name,
                    event_hash=event_hash,
                    event_abi=event_abi,
                    context=context,
                )
            except Exception:
                pass
        else:
            raise Exception('could not find event abi')

    # get contract abi
    if contract_abi is None:
        if contract_address is None:
            raise Exception('must specify contract_abi or contract_address')
        contract_abi = await contract_abi_utils.async_get_contract_abi(
            contract_address=contract_address,
            context=context,
        )

    # get event abi
    try:
        return evm.get_event_abi(
            contract_abi=contract_abi,
            event_name=event_name,
            event_hash=event_hash,
            event_abi=event_abi,
        )

    except LookupError as e:

        from ctc import config

        # query contract_abi again if contract abi might have changed since db
        if contract_address is not None:
            contract_abi = await contract_abi_utils.async_get_contract_abi(
                contract_address=contract_address,
                context=config.update_context(context=context, cache=False)
            )

            return evm.get_event_abi(
                contract_abi=contract_abi,
                event_name=event_name,
                event_hash=event_hash,
                event_abi=event_abi,
            )

        else:
            raise e


def get_event_abis(
    contract_abi: spec.ContractABI,
) -> typing.Sequence[spec.EventABI]:
    """get event ABI's from contract ABI"""
    return [item for item in contract_abi if item['type'] == 'event']



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': ascii_command,
        'help': 'convert hex to ascii',
        'args': [
            {'name': 'data', 'help': 'hex data to convert'},
        ],
        'examples': [
            '0x766974616c696b',
            '0x7475726475636b656e',
        ],
    }


def ascii_command(data: str) -> None:
    text = evm.binary_to_text(data)
    print(text)



from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils
from . import transaction_serialize
from . import transaction_hashes
from . import transaction_types


def sign_transaction(
    transaction: spec.PrechainTransaction,
    *,
    private_key: str,
    chain_id: int | None = None,
) -> tuple[int, int, int]:
    """sign transaction using private key"""

    tx_chain_id: typing.Union[int, bytes, str, None]
    if chain_id is None and 'chain_id' in transaction:
        tx_chain_id = transaction['chain_id']
        if tx_chain_id is not None:
            chain_id = binary_utils.binary_convert(tx_chain_id, 'integer')
    elif chain_id is None and 'chainId' in transaction:
        tx_chain_id = typing.cast(
            typing.Union[int, bytes, str, None], transaction.get('chainId')
        )
        if tx_chain_id is not None:
            chain_id = binary_utils.binary_convert(tx_chain_id, 'integer')
    message = transaction_serialize.serialize_unsigned_transaction(
        transaction,
        chain_id=chain_id,
    )
    return binary_utils.sign_data_message(
        message=message,
        private_key=private_key,
        chain_id=chain_id,
        mode='eth_sign',
    )


def verify_transaction_signature(
    *,
    transaction: spec.PrechainTransaction,
    signature: spec.Signature | None = None,
    public_key: spec.Data | None = None,
    address: spec.Data | None = None,
) -> bool:
    """verify that transaction was signed by given public key"""

    # extract signature from transaction
    if signature is None:
        if (
            transaction.get('v') is not None
            and transaction.get('r') is not None
            and transaction.get('s') is not None
        ):
            signature = (transaction['v'], transaction['r'], transaction['s'])
        else:
            raise Exception('must provide signature for transaction')

    # get transaction hash
    transaction_hash = transaction_hashes.hash_unsigned_transaction(transaction)

    # process signature
    v, r, s = binary_utils.unpack_signature_vrs(signature)
    type = transaction_types.get_transaction_type(transaction)
    if type == 2 and v in [0, 1]:
        v = v + 27

    # verify signature
    return binary_utils.verify_signature(
        signature=(v, r, s),
        message_hash=transaction_hash,
        public_key=public_key,
        address=address,
    )


def recover_transaction_sender(
    transaction: spec.PrechainTransaction,
    signature: spec.Signature,
) -> spec.Address:
    """recover signing address of transaction from signature

    adapted from https://github.com/ethereum/pyethereum/blob/ecb14c937a0b6cb0a0dc4f06be3a88e6d53dcce3/ethereum/transactions.py#L68
    """

    if 'from' in transaction:
        return transaction['from']  # type: ignore

    v, r, s = binary_utils.unpack_signature_vrs(signature)

    # encode transaction data
    if r == 0 and s == 0:
        # null address
        return '0x' + '0' * 40
    elif v in (27, 28):
        chain_id = None
    elif v >= 37:
        chain_id = binary_utils.get_signature_network_id(signature)
        if chain_id is None:
            raise Exception('could not parse chain_id')
        v = v - chain_id * 2 - 8
        assert v in (27, 28)
    else:
        raise Exception('invalid v value')
    if (
        r >= binary_utils.signature_utils.secp256k1_utils.N
        or s >= binary_utils.signature_utils.secp256k1_utils.N
        or r == 0
        or s == 0
    ):
        raise Exception('invalid signature values')

    # get public key
    rlp_transaction = transaction_serialize.serialize_unsigned_transaction(
        transaction,
        chain_id=chain_id,
    )
    message_hash = binary_utils.keccak(rlp_transaction)
    public_key = binary_utils.recover_signer_public_key(
        message_hash=message_hash,
        signature=(v, r, s),
    )

    # convert public key to address
    sender = '0x' + binary_utils.keccak(public_key)[-40:]
    return sender


def is_transaction_signed(transaction: typing.Mapping[str, typing.Any]) -> bool:
    """return whether transaction is signed"""

    return (
        transaction.get('v') is not None
        and transaction.get('r') is not None
        and transaction.get('s') is not None
    )




from __future__ import annotations

import ast
import json
import os
import subprocess
import tempfile
import typing

import pytest

import ctc
from ctc import spec
from ctc.config import config_defaults
from ctc.config import upgrade_utils


#
# # utility functions
#


def run_ctc_command(
    subcommand: str | typing.Sequence[str],
    max_time: int | None = 30,
    **cli_args: typing.Any,
) -> str:

    if isinstance(subcommand, str):
        subcommand = [subcommand]

    # assemble command pieces
    command_pieces = ['python3', '-m', 'ctc'] + list(subcommand)
    for key, value in cli_args.items():

        if value is False:
            continue

        command_pieces.append('--' + key.replace('_', '-'))
        if isinstance(value, str):
            command_pieces.append(value)
        elif value is True:
            pass
        elif isinstance(value, int):
            command_pieces.append(str(value))
        else:
            raise Exception('arg values should be True or str')
    command_pieces = [
        piece.strip('"') for piece in command_pieces if piece != ''
    ]

    # run command
    print('running command: ' + ' '.join(command_pieces))
    output = subprocess.check_output(
        command_pieces, timeout=max_time, env=os.environ.copy()
    )
    return output.decode().rstrip('\n')


def run_ctc_headless(
    max_time: int | None = 30,
    ignore_old_config=True,
    skip_aliases=True,
    **cli_args: typing.Any,
) -> typing.Any:

    return run_ctc_command(
        subcommand='setup',
        max_time=max_time,
        headless=True,
        ignore_old_config=ignore_old_config,
        skip_aliases=skip_aliases,
        **cli_args,
    )


def get_ctc_config() -> typing.Mapping[typing.Any]:
    output = run_ctc_command('config', json=True)
    return ast.literal_eval(output)


def get_default_provider() -> spec.Provider:
    config = get_ctc_config()
    default_network = config['default_network']
    default_provider = config['default_providers'][default_network]
    return config['providers'][default_provider]


def get_rpc_urls():
    config = get_ctc_config()
    return [provider['url'] for provider in config['providers'].values()]


def create_temp_config_path():
    tempdir = tempfile.mkdtemp()
    return os.path.join(tempdir, 'config.json')


def get_db_path():
    config = get_ctc_config()
    return config['db_configs']['main']['path']


#
# # tests
#


def test_use_env_var_config_path(monkeypatch):
    # test that ctc reads from the CTC_CONFIG_PATH env var
    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)
    actual = run_ctc_command(['config', 'path'])
    assert config_path == actual


def test_ctc_setup__env_var_sets_config_path(monkeypatch):
    # test that ctc writes to the CTC_CONFIG_PATH env var
    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)
    assert not os.path.isfile(config_path)
    run_ctc_headless(skip_networking=True)
    assert os.path.isfile(config_path)


def test_ctc_setup__requires_rpc(monkeypatch):
    # test that ctc runs into headless exception
    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)
    if 'ETH_RPC_URL' in os.environ:
        monkeypatch.delenv('ETH_RPC_URL')
    with pytest.raises(Exception):
        run_ctc_headless(skip_db=True)


def test_ctc_setup__with_env_eth_rpc_url(monkeypatch):
    # test that ctc setup can use ETH_RPC_URL

    # requires internet to run properly

    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)

    test_url = 'https://some/test/url'

    # set rpc in env var
    monkeypatch.setenv('ETH_RPC_URL', test_url)
    monkeypatch.setenv('ETH_RPC_CHAIN_ID', '3')

    # perform setup
    run_ctc_headless(skip_db=True)

    # ensure that newly written config file has specified rpc
    default_provider = get_default_provider()
    assert test_url == default_provider['url']
    assert 3 == default_provider['network']


def test_ctc_setup__with_rpc_command(monkeypatch):
    # test that ctc setup can use rpc-url

    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)

    test_url = 'https://some/test/url'
    test_chain_id = '10'

    # perform setup
    if 'ETH_RPC_URL' in os.environ:
        monkeypatch.delenv('ETH_RPC_URL')
    run_ctc_headless(
        rpc_url=test_url,
        rpc_chain_id=test_chain_id,
        skip_db=True,
    )

    # ensure that newly written config file has specified rpc
    default_provider = get_default_provider()
    assert test_url == default_provider['url']
    assert int(test_chain_id) == default_provider['network']


def test_ctc_setup__set_data_dir(monkeypatch):

    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)

    data_dir = tempfile.mkdtemp()
    assert len(os.listdir(data_dir)) == 0
    run_ctc_headless(
        data_dir=data_dir,
        rpc_url='https://some/test/url',
        rpc_chain_id=1,
    )
    assert len(os.listdir(data_dir)) > 0


def test_ctc_setup__skip_db(monkeypatch):
    # test that skipping or not skipping db setup works properly

    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)

    data_dir = tempfile.mkdtemp()
    run_ctc_headless(
        data_dir=data_dir,
        skip_db=True,
        rpc_url='https://some/test/url',
        rpc_chain_id=1,
    )

    db_path = get_db_path()
    assert db_path.startswith(data_dir)
    assert not os.path.isfile(db_path)

    # now rerun without skipping db, and assert that db is created
    run_ctc_headless(
        data_dir=data_dir,
        skip_db=False,
        rpc_url='https://some/test/url',
        rpc_chain_id=1,
    )
    assert os.path.isfile(db_path)


old_config__0_2_10 = {
    'config_spec_version': '0.2.10',
    'data_dir': os.path.expanduser('~/ctc_data'),
    'providers': {
        'example_provider_1': {
            'url': 'https://example_provider.com',
            'name': 'example_provider_1',
            'network': 1,
            'protocol': 'http',
            'session_kwargs': {},
            'chunk_size': None,
            'convert_reverts_to_none': False,
        },
        'example_provider_2': {
            'url': 'http://localhost',
            'name': 'example_provider_2',
            'network': 42161,
            'protocol': 'http',
            'session_kwargs': {},
            'chunk_size': None,
            'convert_reverts_to_none': False,
        },
    },
    'networks': {
        'mainnet': {
            'name': 'mainnet',
            'chain_id': 1,
            'block_explorer': 'etherscan.io',
        },
        'arbitrum': {
            'name': 'arbitrum',
            'chain_id': 42161,
            'block_explorer': 'arbiscan.io',
        },
    },
    'network_defaults': {
        'default_network': 'mainnet',
        'default_providers': {
            'mainnet': 'example_provider_1',
            'arbitrum': 'example_provider_2',
        },
    },
}

old_config__0_3_0 = upgrade_utils.upgrade__0_2_0__to__0_3_0(old_config__0_2_10)

default_config = config_defaults.get_default_config()

old_config_examples = [
    #
    # 0.2.10 style config
    old_config__0_2_10,
    #
    # blank old config
    {},
    #
    # current style default config
    default_config,
    #
    # default config with small changes
    dict(
        default_config,
        data_dir=tempfile.mkdtemp(),
        providers={},
        default_providers={},
    ),
]


@pytest.mark.parametrize('old_config', old_config_examples)
def test_ctc_setup__use_old_config(monkeypatch, old_config):
    # test that
    # 1) values are preserved from old_configs
    # 2) old_configs do not crash the setup

    config_path = create_temp_config_path()
    monkeypatch.setenv('CTC_CONFIG_PATH', config_path)

    # write old config to config path
    assert not os.path.isfile(config_path)
    with open(config_path, 'w') as f:
        json.dump(old_config, f)

    # run setup
    run_ctc_headless(
        skip_networking=True,
        ignore_old_config=False,
        overwrite=True,
        skip_db=True,
    )

    old_config_upgraded = upgrade_utils.upgrade_config(old_config)

    # now, check that old_config values are preserved if they were valid
    new_config = get_ctc_config()

    # check that current ctc value is used
    assert new_config['config_spec_version'] in [
        ctc.__version__,
        upgrade_utils.get_stable_version(ctc.__version__),
    ]

    # check that providers are preserved
    assert len(old_config_upgraded['providers']) == len(new_config['providers'])

    # check that default network is preserved
    assert (
        old_config_upgraded.get('default_network')
        == new_config['default_network']
    )

    # check that default providers are preserved
    assert len(old_config_upgraded['default_providers']) == len(
        new_config['default_providers']
    )
    for key, value in old_config_upgraded['default_providers'].items():
        assert new_config['default_providers'].get(key) == value

    # check that custom networks are preserved
    old_networks = old_config_upgraded.get('networks')
    if old_networks is not None:
        for network_id, network_metadata in old_networks.items():
            assert network_id in new_config['networks']
            for key, value in network_metadata.items():
                assert new_config['networks'][network_id].get(key) == value

    # check that data dir is preserved
    old_data_dir = old_config_upgraded.get('data_dir')
    if old_data_dir is not None:
        assert new_config['data_dir'] == old_data_dir



from __future__ import annotations

import typing

import toolstr

from ctc import spec
from .. import binary_utils
from .. import block_utils
from . import event_query_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def _async_query_events_from_node(
    *,
    contract_address: spec.Address | None,
    event_hash: typing.Any | None,
    topic1: typing.Any | None,
    topic2: typing.Any | None,
    topic3: typing.Any | None,
    start_block: int,
    end_block: int,
    context: spec.Context = None,
    verbose: bool | int,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'binary',
    chunk_size: int = 100000,
    max_blocks_per_request: int = 2000,
) -> spec.DataFrame:
    """query events from node and cache results in db if desired"""

    import asyncio
    import polars as pl
    from ctc.toolbox import pl_utils
    from ctc.toolbox import range_utils

    # parse query type
    query_type = event_query_utils._parse_event_query_type(
        contract_address=contract_address,
        event_hash=event_hash,
        topic1=topic1,
        topic2=topic2,
        topic3=topic3,
    )
    if query_type == 8:
        raise Exception(
            'querying only by non-event-type topics is unsupported by some providers'
        )

    # break into chunks, each will be independently written to db
    chunk_ranges = range_utils.range_to_chunks(
        start=start_block,
        end=end_block,
        chunk_size=chunk_size,
    )

    if verbose >= 1:
        n_blocks = end_block - start_block + 1
        print('fetching events from node over', toolstr.format(n_blocks), 'blocks')
    if verbose >= 2:
        from ctc import cli
        from ctc import config

        event_query_utils.print_event_query_summary(
            contract_address=contract_address,
            event_hash=event_hash,
            topic1=topic1,
            topic2=topic2,
            topic3=topic3,
            start_block=start_block,
            end_block=end_block,
        )
        network = config.get_context_chain_id(context)
        cli.print_bullet(key='network', value=network)
        cli.print_bullet(key='chunk_size', value=chunk_size)
        cli.print_bullet(key='n_chunks', value=len(chunk_ranges))

    # TODO: make this async, store indirect reference in dict and pass dict
    latest_block_number = await block_utils.async_get_latest_block_number(
        context=context
    )

    # process each meta chunk
    coroutines = []
    for chunk_start, chunk_end in chunk_ranges:
        coroutine = _async_query_node_events_chunk(
            contract_address=contract_address,
            event_hash=event_hash,
            topic1=topic1,
            topic2=topic2,
            topic3=topic3,
            chunk_start=chunk_start,
            chunk_end=chunk_end,
            context=context,
            latest_block_number=latest_block_number,
            max_blocks_per_request=max_blocks_per_request,
        )
        coroutines.append(coroutine)
    chunks = await asyncio.gather(*coroutines)

    # package result in dataframe
    result = [response for chunk in chunks for response in chunk]
    columns = event_query_utils.get_event_df_columns(binary_format='prefix_hex')
    df = pl.DataFrame(result, schema=columns)

    # convert binary output columns
    if binary_output_format == 'prefix_hex':
        # already in prefix hex
        pass
    elif binary_output_format == 'binary':
        df = pl_utils.prefix_hex_columns_to_binary(
            df=df,
            columns=['topic1', 'topic2', 'topic3', 'unindexed'],
        )
    else:
        raise Exception('unknown binary_output_format')

    return df


async def _async_query_node_events_chunk(
    *,
    contract_address: spec.Address | None,
    event_hash: bytes | str | None,
    topic1: bytes | str | None,
    topic2: bytes | str | None,
    topic3: bytes | str | None,
    chunk_start: int,
    chunk_end: int,
    context: spec.Context,
    max_blocks_per_request: int = 2000,
    latest_block_number: int | None = None,
) -> typing.Sequence[spec.EncodedEvent]:
    """process a chunk of events from node"""

    import asyncio
    from ctc import config
    from ctc import rpc
    from ctc.toolbox import range_utils

    # break each meta chunk into requests
    chunk_requests = range_utils.range_to_chunks(
        start=chunk_start,
        end=chunk_end,
        chunk_size=max_blocks_per_request,
    )

    # encode topics
    if event_hash is not None:
        event_hash = binary_utils.to_hex(event_hash)
    if topic1 is not None:
        topic1 = binary_utils.to_hex(topic1)
    if topic2 is not None:
        topic2 = binary_utils.to_hex(topic2)
    if topic3 is not None:
        topic3 = binary_utils.to_hex(topic3)

    # assemble topics
    if topic3 is not None:
        topics = [event_hash, topic1, topic2, topic3]
    elif topic2 is not None:
        topics = [event_hash, topic1, topic2]
    elif topic1 is not None:
        topics = [event_hash, topic1]
    else:
        topics = [event_hash]

    # request from node
    coroutines = []
    for request_start, request_end in chunk_requests:
        coroutine = rpc.async_eth_get_logs(
            address=contract_address,
            topics=topics,
            start_block=request_start,
            end_block=request_end,
            context=context,
        )
        coroutines.append(coroutine)
    results = await asyncio.gather(*coroutines)

    # process raw events
    raw_logs = [event for result in results for event in result]
    encoded_events = [
        log[:5] + log[5] + ((None,) * (4 - len(log[5]))) + (log[6],)
        for log in raw_logs
    ]

    # write encoded events to database
    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='events', context=context
    )
    if write_cache:
        from ctc import db

        query_type = event_query_utils._parse_event_query_type(
            contract_address=contract_address,
            event_hash=event_hash,
            topic1=topic1,
            topic2=topic2,
            topic3=topic3,
        )

        query: spec.DBEventQuery = {
            'query_type': query_type,
            'contract_address': contract_address,
            'event_hash': event_hash,
            'topic1': topic1,
            'topic2': topic2,
            'topic3': topic3,
            'start_block': chunk_start,
            'end_block': chunk_end,
        }
        await db.async_intake_encoded_events(
            encoded_events=encoded_events,
            query=query,
            context=context,
            latest_block=latest_block_number,
        )

    return encoded_events


# async def _async_process_raw_node_logs(
#     raw_logs: typing.Sequence[spec.RawLog],
# ) -> typing.Sequence[spec.EncodedEvent]:
#     """convert from raw logs from node into encoded events for db"""
#     for log in raw_logs:
#         event: spec.EncodedEvent = log  # type: ignore
#         event['contract_address'] = log.pop('address')  # type: ignore
#         event['unindexed'] = log.pop('data')  # type: ignore
#         topics = log.pop('topics')  # type: ignore
#         if topics is not None and len(topics) > 0:
#             topic_iter = iter(topics)
#             if len(topics) >= 1:
#                 event['event_hash'] = next(topic_iter)
#             if len(topics) >= 2:
#                 event['topic1'] = next(topic_iter)
#             if len(topics) >= 3:
#                 event['topic2'] = next(topic_iter)
#             if len(topics) >= 4:
#                 event['topic3'] = next(topic_iter)
#     return raw_logs  # type: ignore




from __future__ import annotations

import typing

from ctc import spec
from . import block_time_search
from . import block_time_singular

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def async_get_blocks_of_timestamps(
    timestamps: typing.Sequence[int],
    *,
    block_timestamps: typing.Mapping[int, int] | None = None,
    block_number_array: spec.NumpyArray | None = None,
    block_timestamp_array: spec.NumpyArray | None = None,
    nary: int | None = None,
    cache: block_time_search.BlockTimestampSearchCache | None = None,
    mode: Literal['<=', '>=', '=='] = '>=',
    context: spec.Context = None,
) -> list[int]:
    """search for blocks corresponding to list of timestamps"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_timestamps', context=context
    )

    if block_timestamps is not None or (
        block_number_array is not None and block_timestamp_array is not None
    ):
        import numpy as np

        if mode != '>=':
            raise NotImplementedError()

        if block_timestamp_array is None:
            if block_timestamps is None:
                raise Exception('must specify more arguments')
            block_timestamp_array = np.array(list(block_timestamps.values()))
        if block_number_array is None:
            if block_timestamps is None:
                raise Exception('must specify more arguments')
            block_number_array = np.array(list(block_timestamps.keys()))

        blocks = []
        for timestamp in timestamps:
            block = block_time_singular._get_block_of_timestamp_from_arrays(
                timestamp=timestamp,
                block_timestamp_array=block_timestamp_array,
                block_number_array=block_number_array,
                verbose=False,
            )
            blocks.append(block)

        return blocks

    else:

        # get timestamps form db
        if read_cache:
            from ctc import db

            db_blocks = await db.async_query_timestamps_blocks(
                context=context,
                timestamps=timestamps,
                mode=mode,
            )
            if db_blocks is None:
                db_blocks = [None for timestamp in timestamps]

            # package non-null results
            results: dict[int, int] = {}
            remaining_timestamps: list[int] = []
            for possible_block, timestamp in zip(db_blocks, timestamps):
                if possible_block is None:
                    remaining_timestamps.append(timestamp)
                else:
                    results[timestamp] = possible_block
        else:
            remaining_timestamps = list(timestamps)
            results = {}

        # get timestamps from rpc node
        if len(remaining_timestamps) > 0:
            coroutines = []
            for timestamp in remaining_timestamps:
                coroutine = block_time_singular.async_get_block_of_timestamp(
                    timestamp=timestamp,
                    verbose=False,
                    cache=cache,
                    nary=nary,
                    context=config.update_context(context=context, cache=False),
                    mode=mode,
                )
                coroutines.append(coroutine)
            import asyncio

            node_blocks = await asyncio.gather(*coroutines)
            node_results = dict(zip(remaining_timestamps, node_blocks))
            results.update(node_results)

        # combine
        return [results[timestamp] for timestamp in timestamps]



from __future__ import annotations

import typing

from ctc import spec
from .. import erc20_utils
from . import erc4626_metadata

if typing.TYPE_CHECKING:
    import polars as pl


#
# # normalize shares
#


@typing.overload
async def async_normalize_erc4626_shares(
    token: spec.Address,
    shares: int,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> float:
    ...


@typing.overload
async def async_normalize_erc4626_shares(
    token: spec.Address,
    shares: pl.Series,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> pl.Series:
    ...


@typing.overload
async def async_normalize_erc4626_shares(
    token: spec.Address,
    shares: spec.NumpyArray,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> spec.NumpyArray:
    ...


@typing.overload
async def async_normalize_erc4626_shares(
    token: spec.Address,
    shares: typing.Sequence[int],
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> typing.Sequence[float]:
    ...


async def async_normalize_erc4626_shares(
    token: spec.Address,
    shares: int | typing.Sequence[int] | pl.Series | spec.NumpyArray,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> float | typing.Sequence[float] | pl.Series | spec.NumpyArray:
    """normalize ERC-4626 vault shares (divide by 10 ** decimals)"""

    if hasattr(shares, '__iter__'):
        return await erc20_utils.async_normalize_erc20_quantities(
            quantities=typing.cast(typing.Sequence[int], shares),
            token=token,
            block=block,
            decimals=decimals,
            context=context,
        )

    else:
        return await erc20_utils.async_normalize_erc20_quantity(
            quantity=shares,
            token=token,
            block=block,
            decimals=decimals,
            context=context,
        )


async def async_normalize_erc4626s_shares(
    tokens: typing.Sequence[spec.Address],
    shares: typing.Sequence[int],
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decimals: typing.Optional[typing.Sequence[typing.SupportsInt]] = None,
) -> typing.Sequence[float]:
    """normalize ERC-4626 vaults shares (divide by 10 ** decimals)"""

    return await erc20_utils.async_normalize_erc20s_quantities(
        quantities=shares,
        tokens=tokens,
        block=block,
        decimals=decimals,
        context=context,
    )


#
# # normalize assets
#


@typing.overload
async def async_normalize_erc4626_assets(
    token: spec.Address,
    assets: int,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> float:
    ...


@typing.overload
async def async_normalize_erc4626_assets(
    token: spec.Address,
    assets: pl.Series,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> pl.Series:
    ...


@typing.overload
async def async_normalize_erc4626_assets(
    token: spec.Address,
    assets: spec.NumpyArray,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> spec.NumpyArray:
    ...


@typing.overload
async def async_normalize_erc4626_assets(
    token: spec.Address,
    assets: typing.Sequence[int],
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> typing.Sequence[float]:
    ...


async def async_normalize_erc4626_assets(
    token: spec.Address,
    assets: int | typing.Sequence[int] | pl.Series | spec.NumpyArray,
    *,
    block: spec.BlockNumberReference | None = None,
    decimals: typing.SupportsInt | None = None,
    context: spec.Context = None,
) -> float | typing.Sequence[float] | pl.Series | spec.NumpyArray:
    """normalize ERC-4626 vault assets (divide by 10 ** decimals)"""

    asset = await erc4626_metadata.async_get_erc4626_asset(
        token=token,
        block=block,
        context=context,
    )

    if hasattr(assets, '__iter__'):
        return await erc20_utils.async_normalize_erc20_quantities(
            quantities=typing.cast(typing.Sequence[int], assets),
            token=asset,
            block=block,
            decimals=decimals,
            context=context,
        )
    else:
        return await erc20_utils.async_normalize_erc20_quantity(
            quantity=assets,
            token=asset,
            block=block,
            decimals=decimals,
            context=context,
        )


async def async_normalize_erc4626s_assets(
    tokens: typing.Sequence[spec.Address],
    assets: typing.Sequence[int],
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decimals: typing.Optional[typing.Sequence[typing.SupportsInt]] = None,
) -> typing.Sequence[float]:
    """normalize ERC-4626 vaults assets (divide by 10 ** decimals)"""

    asset_addresses = await erc4626_metadata.async_get_erc4626s_assets(
        tokens=tokens,
        block=block,
        context=context,
    )

    return await erc20_utils.async_normalize_erc20s_quantities(
        quantities=assets,
        tokens=asset_addresses,
        block=block,
        decimals=decimals,
        context=context,
    )



from __future__ import annotations

import toolcli

from .. import llama_tvls


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_command,
        'help': 'output data from Defi Llama api',
        'examples': [''],
    }


async def async_llama_command() -> None:
    await llama_tvls.async_print_historical_defi_tvl()
